"""Methods for estimating and accessing properties of molecules."""
from itertools import product
import re
from typing import Union
import warnings

import numpy as np
import pandas as pd
import periodictable as pt

from carpy.gaskinetics._chemistry_elements import Elements
from carpy.utility import (
    GetPath, LoadData, Quantity, Unicodify,
    call_depth, cast2numpy, cast2quantity, constants as co
)

__all__ = ["Atom", "MoleculeAcyclic"]
__author__ = "Yaseen Reza"

# ============================================================================ #
# Load data required by the module
# ---------------------------------------------------------------------------- #
bond_datatypes = ["forceconstants", "lengths", "strengths"]
dfs_bonds = {
    k: LoadData.yaml(GetPath.localpackage("data", f"bond_{k}.yaml"))
    for k in bond_datatypes
}

dfs_criticalpts = LoadData.excel(
    GetPath.localpackage("data", "molecule_criticalpoints.xlsx"))
criticalpoints = cast2quantity(dfs_criticalpts["main"])


# ============================================================================ #
# Support functions and definitions
# ---------------------------------------------------------------------------- #
class Atom(object):
    """An object for tracking electron coordination."""
    _oxidation = 0
    _valence0: int
    _dvalence_simple = 0
    _dvalence_dative = 0

    def __init__(self, element: Union[pt.core.Element, str], /):
        if isinstance(element, pt.core.Element):
            pt_element = element
        else:
            pt_element = getattr(pt, element)
        self._element = pt_element
        self._period, self._group = Elements.locate(pt_element)
        self._ions = sorted(self.element.ions + (0,))
        self._valence0 = Elements.valence(pt_element)
        return

    def __repr__(self):
        reprstring = f"{type(self).__name__}({self.element})"
        return reprstring

    def __str__(self):
        if self.oxidation == 0:
            oxidation = ""
        elif self.oxidation > 0:
            oxidation = Unicodify.superscript_all(f"{self.oxidation}+")
        else:
            oxidation = Unicodify.superscript_all(f"{abs(self.oxidation)}-")
        rtnstring = f"{self.element.symbol}{oxidation}"
        return rtnstring

    @property
    def element(self) -> pt.core.Element:
        """Object of'periodictable' module, describing the atomic element."""
        return self._element

    @property
    def oxidation(self) -> int:
        """Oxidation state of the atom."""
        return self._oxidation

    @property
    def ions(self) -> list:
        """Possible ionic charges the atom can have."""
        return self._ions

    @property
    def valence(self):
        """Number of outer-shell electrons free for bonding."""
        dvalence = self._dvalence_simple + min(0, self._dvalence_dative)
        return self._valence0 + dvalence

    @property
    def n_bonds(self):
        """Number of bonds that have been made with other atoms."""
        return abs(self._dvalence_simple) + int(abs(self._dvalence_dative / 2))

    def bond_covalent_simple(self, other: object):
        """
        Simple covalent bonding from one Atom object to another Atom object.

        Args:
            other: Another object of this class.

        Returns:
            tuple: own atom, simple covalent bond order, and other atom.

        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"'{other}' was not of type '{type(self)}'")

        # Guess whether or not own atom should be oxidising or reducing
        self_oxidising = None

        # Are the atoms similar?
        if self.element is other.element:
            pass

        # Atoms are dis-similar
        else:
            # Are the atoms already oxidising xor reducing?
            if self.oxidation > 0 or other.oxidation < 0:
                self_oxidising = True
            elif self.oxidation < 0 or other.oxidation > 0:
                self_oxidising = False
            elif self.oxidation > 0 and other.oxidation > 0:
                return None
            elif self.oxidation < 0 and other.oxidation < 0:
                return None

            # No prior oxidation has occurred, force hydrogen to oxidise
            elif self.element is getattr(pt, "H") is not other.element:
                # If it's possible for other to reduce, then it must do so
                if (np.array(other.ions) < 0).any() and other.oxidation <= 0:
                    self_oxidising = True  # H is oxidising (losing electrons)
                else:
                    self_oxidising = False
            elif self.element is not getattr(pt, "H") is other.element:
                # If it's possible for self to reduce, then it must do so
                if (np.array(self.ions) < 0).any() and self.oxidation <= 0:
                    self_oxidising = False  # H is oxidising (losing electrons)
                else:
                    self_oxidising = True

            # Guess from electronegativity
            else:
                chi_self_i = Elements.by_pseudoChi.index(self.element)
                chi_othr_i = Elements.by_pseudoChi.index(other.element)
                if chi_self_i < chi_othr_i:  # self is the oxidising AGENT
                    self_oxidising = False  # self must gain electrons
                elif chi_self_i > chi_othr_i:  # other is the oxidising AGENT
                    self_oxidising = True  # self must lose electrons

        if self_oxidising is False:
            self_dval_limit = abs(min(self.ions)) - self.n_bonds
            othr_dval_limit = abs(min(other.ions)) - other.n_bonds
            if self_dval_limit == 0:
                self_dval_limit = abs(max(self.ions)) - self.n_bonds
            if othr_dval_limit == 0:
                othr_dval_limit = abs(max(other.ions)) - other.n_bonds
        elif self_oxidising is True:
            self_dval_limit = abs(min(self.ions)) - self.n_bonds
            othr_dval_limit = abs(min(other.ions)) - other.n_bonds
            if self_dval_limit == 0:
                self_dval_limit = abs(max(self.ions)) - self.n_bonds
            if othr_dval_limit == 0:
                othr_dval_limit = abs(max(other.ions)) - other.n_bonds
        else:
            # Molecules must be similar. Choose minimum possible dvalence
            self_dval_limit = abs(min(self.ions)) - self.n_bonds
            othr_dval_limit = abs(min(other.ions)) - other.n_bonds
            # ... except for when minimum possible dvalence == 0, take next ions
            if self_dval_limit == 0 and othr_dval_limit == 0:
                self_dval_limit = int((max(self.ions) - self.n_bonds) / 2)
                othr_dval_limit = int((max(other.ions) - other.n_bonds) / 2)
        bond_order = np.abs((self_dval_limit, othr_dval_limit)).min()
        if bond_order == 0:
            return None

        # Update oxidisation states depending on which atom was oxidising
        if self_oxidising is False:
            self._oxidation -= bond_order
            other._oxidation += bond_order
        elif self_oxidising is True:
            self._oxidation += bond_order
            other._oxidation -= bond_order
        # Update valence states
        self._dvalence_simple -= bond_order
        other._dvalence_simple -= bond_order

        return self, bond_order, other

    def bond_covalent_dative(self, other: object):
        """
        Dative covalent bonding from one Atom object to another Atom object.

        Args:
            other: Another object of this class.

        Returns:
            tuple: own atom, dative covalent bond order, and other atom.

        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"'{other}' was not of type '{type(self)}'")

        # Oxidation is irrelevant in this process, since we don't expect changes
        if self.n_bonds < other.n_bonds:
            recipient = self
            donor = other
        elif self.n_bonds > other.n_bonds:
            donor = self
            recipient = other
        else:
            donor = self
            recipient = other

        if donor.valence >= 2:
            donor._dvalence_dative -= 2
            recipient._dvalence_dative += 2
            return self, 1, other

        return None


@call_depth
def get_connectivity(formula: str, output: list):
    """
    Determine the types of covalent bonds present between atoms of a molecule.
    This function behaves unusually, in that the function's output is returned
    via the modification of an empty list provided to it by the user. Don't ask
    me why, I don't know either.

    Args:
        formula: Condensed structural formula of a molecule.
        output: An empty list, into which the output of this function is saved.

    Notes:
        I've not used threading before, but this function relies on being able
        to know how deep in a state of recursion it is. It's possible that a
        `call_depth` attribute may produce unexpected behaviour if parallelised.

    """

    # Replace repeating units with verbose form
    def expand_unit(match_obj):
        """Replace matches representing repeating units with verbose forms."""
        fulltext = match_obj.group()
        sub_unit, n_units = fulltext.rsplit("]", maxsplit=1)
        sub_unit = sub_unit[1:]
        n_units = int(n_units) if n_units.isnumeric() else 1
        return sub_unit * n_units

    formula = re.sub(r"\[.+]\d*", expand_unit, formula)

    # Recursively enter deeper and deeper branches of the molecule
    branches = re.findall(r"\(.+\)\d*", formula)
    atoms = [formula]
    for branch in branches:
        sub_formula, multiple = branch.rsplit(")", maxsplit=1)
        sub_formula = sub_formula[1:]
        multiple = int(multiple) if multiple.isnumeric() else 1
        branch_atoms = []
        for _ in range(multiple):
            branch_atoms += [get_connectivity(sub_formula, output)]
        atoms[-1] = atoms[-1].split(branch, maxsplit=1)
        # noinspection PyTypeChecker
        atoms[-1].insert(-1, branch_atoms)
        # noinspection PyUnresolvedReferences
        atoms = np.array([x for x in pd.core.common.flatten(atoms) if bool(x)])

    # At the deepest un-branched level of the molecule, find the atoms
    for i, atom in enumerate(atoms):
        if isinstance(atom, str):
            atoms[i] = [
                Atom(getattr(pt, symbol))
                for (symbol, mul) in re.findall(
                    rf"({Elements.regex.pattern})(\d*)",
                    atom
                )
                for _ in range(int(mul) if mul.isnumeric() else 1)
            ]
    else:
        # noinspection PyUnresolvedReferences
        atoms = np.array([x for x in pd.core.common.flatten(atoms) if bool(x)])

    # Carry out simple covalent bonding on sub-units
    while True:
        # ----------------------------------------------------------------------
        # Find a candidate for a "central" atom (is able to oxidise/reduce)
        # ----------------------------------------------------------------------
        indices = np.arange(len(atoms))

        central_atoms = []
        bonding_atoms = []
        for atom in atoms:
            # Could the atom be a "central" atom?
            if atom.valence <= 1 or atom.oxidation != 0:
                central_atoms.append(0)
            else:
                central_atoms.append(1)

            # Could the atom be a "subsidiary" bonding atom?
            if atom.valence == 0:
                bonding_atoms.append(0)
            elif atom.oxidation == 0 and atom.n_bonds == abs(atom.ions[0]):
                bonding_atoms.append(0)
            elif atom.oxidation not in [0] + atom.ions[1:-1]:
                bonding_atoms.append(0)
            else:
                bonding_atoms.append(1)
        else:
            central_atoms = cast2numpy(central_atoms)
            bonding_atoms = cast2numpy(bonding_atoms)

            # A central atom must be able to bond to things!
            central_atoms = central_atoms & bonding_atoms
            if np.sum(central_atoms) == 0:
                break

            # There must be other atoms to bond to!
            central_i = np.argmax(central_atoms)
            subsdry_i = indices[(indices != central_i) & (bonding_atoms == 1)]
            if np.sum(subsdry_i) == 0:
                break

        # ----------------------------------------------------------------------
        # Iterate over subsidiary atoms and bond as many times as possible
        # ----------------------------------------------------------------------
        central = atoms[central_i]

        for subsidiary in atoms[subsdry_i]:
            covalent_simple = central.bond_covalent_simple(other=subsidiary)
            if covalent_simple is not None:
                output.append(covalent_simple)
            else:
                continue

        continue

    # If we're in a branch (recursive depth > 0), ascend the recursive stack
    if get_connectivity.call_depth > 0:
        return atoms

    # --------------------------------------------------------------------------
    # With basic bonds completed, we need to look at unusual scenarios next
    # --------------------------------------------------------------------------
    # For monatomic molecules... there are no bonds (to be made)!
    if len(atoms) == 1:
        output.append((atoms[0], 0, None))
        return None

    # For diatomic molecules...
    elif len(atoms) == 2:

        # If homonuclear...
        if atoms[0].element == atoms[1].element:

            # If there are no bonds (invalid central atom candidates earlier)
            if len(output) == 0:
                # Spawn a simple covalent bond
                covalent_simple = atoms[0].bond_covalent_simple(other=atoms[1])
                if covalent_simple is not None:
                    output.append(covalent_simple)

            return None

        # Else, attempt dative covalent bonding
        warnmsg = (
            "Definitely an issue here with forcibly datively covalent bonding, "
            "e.g. that in IBr (dissimilar elements with valence > 2). Need to "
            "add a check that prevents bonding if maximum shell capacity is met"
        )
        warnings.warn(message=warnmsg, category=RuntimeWarning)
        covalent_dative = atoms[0].bond_covalent_dative(other=atoms[1])
        if covalent_dative is not None:
            replace_i, bond_new = \
                ([
                     (i, (a, o + 1, b))
                     for i, (a, o, b) in enumerate(output)
                     if a in covalent_dative and b in covalent_dative
                 ] + [(-1, covalent_dative)])[0]
            if replace_i != -1:
                output[replace_i] = bond_new
            else:
                output.append(bond_new)
            return None

    # Perform depth-first-search to determine if graph is fully connected
    mat_adjacency = np.zeros((len(atoms), len(atoms)), dtype=np.int32)
    atompairs = \
        [tuple(np.where(atoms == x) for x in (b, a)) for (a, _, b) in output]
    for atompair in atompairs:
        mat_adjacency[atompair] = mat_adjacency[atompair[::-1]] = 1

    # Dictionary comprehension to identify nodes (k) and neighbours (v)
    graph = {
        component: atoms.take(indices[mat_adjacency[i] == 1])
        for i, component in enumerate(atoms)
    }

    # If graph is disconnected, we'll know how using a depth first search
    def dfs(query: Atom):
        """Using depth-first search, find networked atoms."""
        # https://medium.com/geekculture/depth-first-search-dfs-algorithm-with-python-2809866cb358
        # Empty set and list to record journey
        visited = set()
        dfs_traversal = list()

        # noinspection PyShadowingNames
        def _dfs(graph, source, visited, dfs_traversal):
            """Depth first search"""
            if source not in visited:
                dfs_traversal.append(source)
                visited.add(source)

                for neighbor_node in graph[source]:
                    _dfs(graph, neighbor_node, visited, dfs_traversal)

            return visited

        return _dfs(graph, query, visited, dfs_traversal)

    def get_graphs():
        """
        For each atom (node), perform DFS and sort by hash to get (sub-)graphs.
        This would produce repeating graphs, so use sets to squash results.
        Visited nodes are hash-sorted, tuple-cast (hashable type), then set-cast

        Returns:
            Set of unique subgraphs identified in the molecule strucutre

        """
        graphs_by_node = \
            [tuple(sorted(dfs(x), key=lambda y: y.__hash__())) for x in atoms]
        return set(graphs_by_node)

    unique_graphs = get_graphs()

    # The graph is already complete
    if len(unique_graphs) == 1:
        return None

    # As a last ditch effort
    if len(unique_graphs) == 2:

        # 1st: Generate every single simple covalent bonding possibility
        atoms_by_increasing_Chi = [sorted(
            x, key=lambda y: Elements.by_pseudoChi.index(y.element)
        )[::-1] for x in unique_graphs]
        combinations = list(product(*atoms_by_increasing_Chi))

        # 2nd: attempt simple covalent bonding
        for (atom1, atom2) in combinations:
            covalent_simple = atom1.bond_covalent_simple(other=atom2)
            if covalent_simple is not None:
                output.append(covalent_simple)
                return None

        # 3rd: attempt dative (coordinate) covalent bonding
        for (atom1, atom2) in combinations:
            covalent_dative = atom1.bond_covalent_dative(other=atom2)
            if covalent_dative is not None:
                replace_i, bond_new = \
                    ([
                         (i, (a, o + 1, b))
                         for i, (a, o, b) in enumerate(output)
                         if a in covalent_dative and b in covalent_dative
                     ] + [(-1, covalent_dative)])[0]
                if replace_i != -1:
                    output[replace_i] = bond_new
                else:
                    output.append(bond_new)
                return None

    if len(get_graphs()) > 1:
        errormsg = (
            f"Unable to fully resolve the bonds in {formula} "
            f"(got bonds={output} and disconnected subgraphs={get_graphs()}"
        )
        raise ValueError(errormsg)

    return None


# ============================================================================ #
# Molecules
# ---------------------------------------------------------------------------- #

class MoleculeAcyclic(object):
    """
    Guess properties of non-ionic, acyclic (open-chain) compounds from formula.

    Examples: ::

        # Instantiate a molecule using condensed structural formula,
        # e.g. 'CH3[CH2]2CH3' is butane, and 'CH3C(CH3)3' is isobutane.
        >>> butane = MoleculeAcyclic("CH3[CH2]2CH3")

    """

    def __init__(self, formula: str):
        """
        Args:
            formula: Condensed structural formula of an open-chain compound.
                 Atoms enclosed in parentheses are treated as branches off of
                 the main tree of atoms, and those in square brackets are
                 handled as repeating units.
        """
        self._formula = formula

        # VVV Below bonding algorithm is broken, do not use publicly!!!
        # Looks weird, but I promise this really does assign bonds to self
        self._bondtuples = []
        get_connectivity(self._formula, self._bondtuples)

        return

    def __repr__(self):
        reprstring = f"{type(self).__name__}(formula='{self._formula}')"
        return reprstring

    def __str__(self):
        return Unicodify.chemical_formula(self._formula)

    @property
    def atoms(self) -> set[Atom]:
        """Atomic composition of the molecule."""
        atom_pairs = [(a, b) for (a, _, b) in self._bondtuples]
        # noinspection PyUnresolvedReferences
        atoms = set(
            (x for x in pd.core.common.flatten(atom_pairs) if x is not None))
        return atoms

    @property
    def elements(self) -> set[pt.core.Element]:
        """Elemental composition of the molecule."""
        elements = set([atom.element for atom in self.atoms])
        return elements

    @property
    def atom_count(self) -> dict[pt.core.Element, int]:
        """Number of atoms of each unique element in the molecule."""
        # Find the element of each atom present in the molecule
        atom_elements = [atom.element for atom in self.atoms]
        count = {element: 0 for element in set(atom_elements)}
        # Iteratively sum up contributions of each atom to an elemental total
        for atom_element in atom_elements:
            count[atom_element] += 1
        return count

    @property
    def atom_AXE(self) -> dict[Atom, dict]:
        """
        Returns an element agnostic designation of the shape and outer-electron
        configuration of an atom (as in the AXE method of VSEPR theory).
        """
        # Count ligands (atoms bonded to a central atom)
        ligands = {atom: list() for atom in self.atoms}
        for bond in self._bondtuples:
            atom_a, order, atom_b = bond
            if order == 0:
                continue
            ligands[atom_a] += [atom_b]
            ligands[atom_b] += [atom_a]

        axe = {atom: dict() for atom in self.atoms}
        for atom in axe.keys():
            axe[atom] = dict([
                ("X", ligands[atom]),  # Record ligands
                ("E", int(atom.valence / 2))  # Record electron pairs
            ])

        return axe

    def formula_condensed_structural(self, pretty: bool = None) -> str:
        """
        Structural formula of the compound, as was provided in instantiation.

        Args:
            pretty: Parameter that decides whether the output string format is
                augmented with unicode subscript characters. Optional, defaults
                to False.

        Returns:
            String representing the molecule's condensed structural formula.

        """
        # Recast as necessary
        pretty = False if pretty is None else pretty

        # Recast as necessary
        if pretty is True:
            formula = Unicodify.chemical_formula(self._formula)
        else:
            formula = self._formula

        return formula

    def formula_empirical(self, pretty: bool = None) -> str:
        """
        Empirical formula of the compound according to IUPAC recommendations.

        Args:
            pretty: Parameter that decides whether the output string format is
                augmented with unicode subscript characters. Optional, defaults
                to False.

        Returns:
            String representing the molecule's empirical formula.

        Notes:
            From "Principles of Chemical Nomenclature", a guide to IUPAC
            recommendations, by G.J. Leigh, H.A. Favre, and W.V. Metanomski.

        """
        # Recast as necessary
        pretty = False if pretty is None else pretty

        # noinspection PyUnresolvedReferences
        gcd = np.gcd.reduce(cast2numpy(list(self.atoms.values())))
        groups = {k.symbol: int(v / gcd) for (k, v) in self.atom_count.items()}

        # If the molecule is organic, sort with C and H first
        indexing_pattern = "".join([f"{x}," for x in Elements.by_alpha])
        if "C" in groups:
            indexing_pattern = "C,H," + indexing_pattern

        # Sort elemental symbols (with C,H if necessary and then) alphabetically
        ordered_keys = \
            sorted(groups, key=lambda x: indexing_pattern.index(f"{x},"))
        formula = "".join([
            f"{symbol}{groups[symbol]}" if groups[symbol] != 1 else f"{symbol}"
            for symbol in ordered_keys
        ])

        # Recast as necessary
        if pretty is True:
            formula = Unicodify.chemical_formula(formula)

        return formula

    def formula_molecular(self, pretty: bool = None) -> str:
        """
        Molecular formula of the compound according to IUPAC recommendations.

        Args:
            pretty: Parameter that decides whether the output string format is
                augmented with unicode subscript characters. Optional, defaults
                to False.

        Returns:
            String representing the molecule's molecular formula.

        Notes:
            From "Principles of Chemical Nomenclature", a guide to IUPAC
            recommendations, by G.J. Leigh, H.A. Favre, and W.V. Metanomski.

        """
        # Recast as necessary
        pretty = False if pretty is None else pretty
        groups = {k.symbol: v for (k, v) in self.atom_count.items()}

        # If the molecule is organic, sort with C and H first
        indexing_pattern = "".join([f"{x}," for x in Elements.by_Z])
        if "C" in groups:
            indexing_pattern = "C,H," + indexing_pattern

        # Sort elemental symbols (with C,H if necessary and then) alphabetically
        ordered_keys = \
            sorted(groups, key=lambda x: indexing_pattern.index(f"{x},"))
        formula = "".join([
            f"{symbol}{groups[symbol]}" if groups[symbol] != 1 else f"{symbol}"
            for symbol in ordered_keys
        ])

        # Recast as necessary
        if pretty is True:
            formula = Unicodify.chemical_formula(formula)

        return formula

    @property
    def critical_point(self) -> tuple:
        """Critical temperature and critical pressure, if known."""
        match = criticalpoints["Molecular Formula"] == self.formula_molecular()
        if any(match):
            index = np.argmax(match)
            # Indexing trick: using a slice retains the units of the Quantity
            T_c = criticalpoints["Critical temperature, T"][index: index + 1]
            p_c = criticalpoints["Critical pressure, p"][index: index + 1]
        else:
            T_c, p_c = np.nan, np.nan

        return T_c, p_c

    @property
    def m(self) -> Quantity:
        """Molecular mass."""
        relative_formula_mass = sum([
            element.mass * num
            for (element, num) in self.atom_count.items()
        ])
        return Quantity(relative_formula_mass, "Da")

    @property
    def M(self) -> Quantity:
        """Molar mass."""
        molar_mass = self.m * co.PHYSICAL.N_A
        return molar_mass

    @property
    def Rs(self) -> Quantity:
        """Specific gas constant."""
        return co.PHYSICAL.R / self.M



"""
Consolidated aircraft recipes in Python (carpy)

This software package contains a variety of tools for the conceptual analysis
of aircraft in handy, pre-packaged modules. Check out the GitHub page for more
information, examples, and tutorials on usage.

carpy is distributed under the GNU GPLv3 License. A full copy of the license
should be present alongside the source code on GitHub.
"""
# The version number of the "entire" package is specified here (for now).
# In future, the idea would be this is tied to the git release (somehow).
__version__ = "0.0.3"

# ============================================================================ #
# -------------------------------- LICENSING --------------------------------- #
# ============================================================================ #
__license__ = "GNU GPLv3"
# Console interaction copyright notice
__copyright__ = (
    f"\nADRpy  Copyright (C) 2023  Yaseen Reza, Luana Defourny"
    f"\nThis program comes with ABSOLUTELY NO WARRANTY; "
    f"for details type `ADRpy.w()'."
    f"\nThis is free software, and you are welcome to redistribute it "
    f"under certain conditions; type `ADRpy.c()' for details."
)


def c() -> None:
    """Redistribution information"""
    disttext = (
        "This program is free software: you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by\n"
        "the Free Software Foundation, either version 3 of the License, or\n"
        "(at your option) any later version."
    )
    print(disttext)
    return None


def w() -> None:
    """Warranty information"""
    warrantytxt = (
        "This program is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU General Public License for more details."
    )
    print(warrantytxt)
    return None

# ============================================================================ #



"""Module for accessing properties of airports and their runways."""
import os
from functools import wraps
import warnings

import numpy as np
import pandas as pd

from carpy.utility import GetPath, Quantity, isNone

__all__ = ["Airport", "Runway"]
__author__ = "Yaseen Reza"


# ============================================================================ #
# Load data required by the module
# ---------------------------------------------------------------------------- #


def load_ourairports(filename: str, /):
    """
    Load and return an offline image of data relating to airfields.

    Args:
        filename: The filename and extension to be loaded.

    References:
        https://ourairports.com/
        https://ourairports.com/help/data-dictionary.html

    Notes:
        Data is offline (imaged), and so may not be up to date!.

    """
    # Load the actual data
    oa_data = os.path.join(GetPath.localpackage(), "data", "ourairports")
    filepath = os.path.join(oa_data, filename)
    dataframe = pd.read_csv(filepath_or_buffer=filepath)

    # Load the definitions (if available)
    try:
        sheetname, _ = filename.rsplit(".")
        filepath = os.path.join(oa_data, "definitions.xlsx")
        definitions = pd.read_excel(io=filepath, sheet_name=sheetname)
    except Exception as _:
        errormsg = (
            f"Could not find definitions for 'ourairports' data: {filename}"
        )
        raise ValueError(errormsg)

    # Package the data and definitions
    classargs = (
        f"{sheetname}data",
        (object,),
        dict([("df", dataframe), ("defs", definitions)])
    )
    dataobject = type(*classargs)

    return dataobject


oa_airports = load_ourairports("airports.csv")
oa_runways = load_ourairports("runways.csv")
warnings.warn(
    message="Using local image of OurAirports data, dated 22/02/2023.",
    category=RuntimeWarning
)


# ============================================================================ #
# Support functions
# ---------------------------------------------------------------------------- #

def idx0(func):
    """
    Decorator for returning the first element of a series returned by the
    wrapped function.

    Args:
        func: Function returning a Pandas series.

    Returns:
        The first element of the series returned by the wrapped function.

    Notes:
        Can also find the first index of an array, but this use case isn't
            officially supported for users - internal use only at this time.

    """
    if callable(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            """Function wrapper, giving the 1st element in a pd.series obj."""
            result = func(*args, **kwargs)
            if isinstance(result, pd.Series):
                return result.iloc[0]
            else:
                return result[0]

        return wrapper

    # Technically the below are exceptions for unexpected uses...
    elif isinstance(func, pd.Series):
        # Oops, not a function returning a series. It's an actual series!
        return func.iloc[0]

    else:
        # Who knows. Just index 0 and hope for the best
        return func[0]


# ============================================================================ #
# Airfield objects
# ---------------------------------------------------------------------------- #


class Runway(object):
    """
    Runway objects describe properties of a runway surface including the
    runway's names (in both directions), position of each runway end, and
    dimensions.
    """

    def __init__(self, ourairports_id: int):
        """
        Args:
            ourairports_id: Numerical ID used by OurAirports internal database.
        """
        self._df = oa_runways.df.loc[oa_runways.df["id"] == ourairports_id]
        return

    def __repr__(self):
        my_id = idx0(self._df["id"])
        return f"{type(self).__name__}(ourairports_id={my_id})"

    def __str__(self):
        return f"RWY {self.name}"

    @property
    @idx0
    def airport_ident(self) -> str:
        """
        The text identifier used in the OurAirports URL of the parent airport.
        This will be the ICAO code if available. Otherwise, it will be a local
        airport code (if no conflict), or if nothing else is available, an
        internally-generated code starting with the ISO2 country code, followed
        by a dash and a four-digit number.
        """
        return self._df["airport_ident"]

    @property
    def name(self) -> str:
        """
        The name of the runway.

        Returns:
            <low-end identifier>/<high-end identifier>

        """
        return f"{idx0(self._df['le_ident'])}/{idx0(self._df['he_ident'])}"

    @property
    def length(self) -> Quantity:
        """
        Length of the full runway surface (including displaced thresholds,
        overrun areas, etc).
        """
        # Convert ourairports data (in feet) to Quantity objects
        return Quantity(idx0(self._df["length_ft"]), "ft")

    @property
    def width(self) -> Quantity:
        """Width of the runway surface."""
        # Convert ourairports data (in feet) to Quantity objects
        return Quantity(idx0(self._df["width_ft"]), "ft")

    @property
    @idx0
    def _surface(self) -> str:
        """Runway surface type."""
        warnmsg = (
            "The '._surface' attribute of runways doesn't adhere to a "
            "controlled vocabulary, therefore code utilising this attribute "
            "should be built and used with caution"
        )
        warnings.warn(message=warnmsg, category=RuntimeWarning)
        return self._df["surface"]

    @property
    def lighted(self) -> bool:
        """True if the surface is lighted at night, else False."""
        return bool(idx0(self._df["lighted"]))

    @property
    def closed(self) -> bool:
        """True if the runway surface is closed, else False."""
        return bool(idx0(self._df["closed"]))

    @property
    @idx0
    def le_ident(self) -> str:
        """Identifier for the low-numbered end of the runway."""
        return self._df["le_ident"]

    @property
    @idx0
    def le_latitude_deg(self) -> float:
        """
        Latitude of the centre of the low-numbered end of the runway, in
        decimal degrees (+ve is north), if available.
        """
        return self._df["le_latitude_deg"]

    @property
    @idx0
    def le_longitude_deg(self) -> float:
        """
        Longitude of the centre of the low-numbered end of the runway, in
        decimal degrees (+ve is east), if available."""
        return self._df["le_longitude_deg"]

    @property
    def le_elevation(self) -> Quantity:
        """
        Elevation above mean sea level of the low-numbered end of the runway.
        """
        return Quantity(idx0(self._df["le_elevation_ft"]), "ft")

    @property
    @idx0
    def le_heading_degT(self) -> float:
        """
        Heading of the low-numbered end of the runway in degrees true (not
        magnetic).
        """
        return self._df["le_heading_degT"]

    @property
    def le_displaced_threshold(self) -> Quantity:
        """
        Length of the displaced threshold (if any) for the low-numbered end of
        the runway. Displaced thresholds may be used for taxiing, takeoff, and
        landing rollout, but not for touchdown.
        """
        displacement = idx0(self._df["le_displaced_threshold_ft"])

        # If there is no displacement, take default value of zero (displacement)
        if np.isnan(displacement):
            displacement = 0

        return Quantity(displacement, "ft")

    @property
    @idx0
    def he_ident(self) -> str:
        """Identifier for the high-numbered end of the runway."""
        return self._df["he_ident"]

    @property
    @idx0
    def he_latitude_deg(self) -> float:
        """
        Latitude of the centre of the high-numbered end of the runway, in
        decimal degrees (+ve is north), if available.
        """
        return self._df["he_latitude_deg"]

    @property
    @idx0
    def he_longitude_deg(self) -> float:
        """
        Longitude of the centre of the high-numbered end of the runway, in
        decimal degrees (+ve is east), if available."""
        return self._df["he_longitude_deg"]

    @property
    def he_elevation(self) -> Quantity:
        """
        Elevation above mean sea level of the high-numbered end of the runway.
        """
        return Quantity(idx0(self._df["he_elevation_ft"]), "ft")

    @property
    @idx0
    def he_heading_degT(self) -> float:
        """
        Heading of the high-numbered end of the runway in degrees true (not
        magnetic).
        """
        return self._df["he_heading_degT"]

    @property
    def he_displaced_threshold(self) -> Quantity:
        """
        Length of the displaced threshold (if any) for the high-numbered end of
        the runway. Displaced thresholds may be used for taxiing, takeoff, and
        landing rollout, but not for touchdown.
        """
        displacement = idx0(self._df["he_displaced_threshold_ft"])

        # If there is no displacement, take default value of zero (displacement)
        if np.isnan(displacement):
            displacement = 0

        return Quantity(displacement, "ft")


class Airport(object):
    """
    Airport objects are used to quickly access size, position, and orientation
    of available runways.
    """

    def __init__(self, ident: str = None, ourairports_id: int = None):
        """
        Args:
            ident: The text identifier used in the OurAirports URL. This will be
                the ICAO code if available. Otherwise, it will be a local
                airport code (if no conflict), or if nothing else is available,
                an internally-generated (by OurAirports) code starting with the
                ISO2 country code, followed by a dash and a four-digit number.
            ourairports_id: Numerical ID used by OurAirports internal database.
        """
        # Sanitise input
        if all(isNone(ident, ourairports_id)):
            errormsg = (
                f"Missing arguments, please supply one of 'ident' or "
                f"'ourairports_id'."
            )
            raise ValueError(errormsg)
        elif ident and ourairports_id:
            errormsg = (
                f"Too many arguments, please supply only one of 'ident' or "
                f"'ourairports_id'."
            )
            raise ValueError(errormsg)

        # Uniquely ID the airport
        if ident:
            df = oa_airports.df.loc[oa_airports.df["ident"] == ident]
        else:
            df = oa_airports.df.loc[oa_airports.df["id"] == ourairports_id]
            ident = df["ident"]
        self._df = df

        # List unique runway surfaces
        rwys_raw = oa_runways.df.loc[oa_runways.df["airport_ident"] == ident]
        self._runways = [Runway(row["id"]) for _, row in rwys_raw.iterrows()]
        return

    def __repr__(self):
        return f"{type(self).__name__}(ident={self.ident})"

    def __str__(self):
        return f"{self.name}"

    @property
    @idx0
    def ident(self) -> str:
        """
        The text identifier used in the OurAirports URL. This will be the ICAO
        code if available. Otherwise, it will be a local airport code (if no
        conflict), or if nothing else is available, an internally-generated code
        starting with the ISO2 country code, followed by a dash and a four-digit
        number.
        """
        return self._df["ident"]

    @property
    @idx0
    def name(self) -> str:
        """The official airport name."""
        return self._df["name"]

    @property
    @idx0
    def latitude_deg(self) -> float:
        """The airport latitude in decimal degrees (+ve for north)."""
        return self._df["latitude_deg"]

    @property
    @idx0
    def longitude_deg(self) -> float:
        """The airport longitude in decimal degrees (+ve for east)."""
        return self._df["longitude_deg"]

    @property
    def elevation(self) -> Quantity:
        """The airport elevation above mean sea level."""
        # Convert ourairports data (in feet) to Quantity objects
        newobject = Quantity(idx0(self._df["elevation_ft"]), "ft")
        return newobject

    @property
    def runways(self) -> list:
        """A list of runway surfaces present at the airport."""
        return self._runways



"""A module of maths utilities."""
import numpy as np

from carpy.utility._miscellaneous import Hint, cast2numpy

__all__ = ["interp_lin", "interp_exp", "moving_average", "point_diff",
           "point_curvature"]
__author__ = "Yaseen Reza"


def interp_lin(x: Hint.nums, xp: Hint.nums, fp: Hint.nums,
               bounded: bool = True) -> Hint.nums:
    """
    A linear interpolation function offering unbounded extrapolation.

    Args:
        x: The values which should be fed to the interpolation function.
        xp: Sorted list of x-values to draw interpolated input from.
        fp: List of return values to draw interpolated outputs from.
            bounded: Boolean, selects whether output is bounded by 'fp' values.
        bounded: If true, output is limited by fp limits. Otherwise extrapolate.

    Returns:
        np.ndarray: The linearly-interpolated values.

    """
    # Recast as necessary
    x = cast2numpy(x)
    xp = cast2numpy(xp)
    fp = cast2numpy(fp)

    # Prepare differentials between each point
    dfp = np.diff(fp)
    dxp = np.diff(xp)

    out = np.zeros_like(fp, shape=x.shape)  # Shape of x, object style of fp

    for i in range(len(dfp)):
        slice = (xp[i] <= x) & (x <= xp[i + 1])
        out[slice] = (x[slice] - xp[i]) * (dfp[i] / dxp[i]) + fp[i]

    sliceL, sliceR = x < xp[0], x > xp[-1]
    if bounded is True:
        out[sliceL] = fp[0]
        out[sliceR] = fp[-1]
    else:
        out[sliceL] = (x[sliceL] - xp[0]) * (dfp / dxp)[0] + fp[0]
        out[sliceR] = (x[sliceR] - xp[-1]) * (dfp / dxp)[-1] + fp[-1]

    return out


def interp_exp(x: Hint.nums, xp: Hint.nums, fp: Hint.nums,
               bounded: bool = True) -> Hint.nums:
    """
    A linear interpolation function offering unbounded extrapolation.

    Args:
        x: The values which should be fed to the interpolation function.
        xp: Sorted list of x-values to draw interpolated input from.
        fp: List of return values to draw interpolated outputs from.
            bounded: Boolean, selects whether output is bounded by 'fp' values.
        bounded: If true, output is limited by fp limits. Otherwise extrapolate.

    Returns:
        np.ndarray: The exponentially-interpolated values.

    """
    # Recast as necessary
    x = cast2numpy(x)
    xp = cast2numpy(xp)
    fp = cast2numpy(fp)

    # Prepare differentials between each point
    log_dfp = np.log(fp[1:] / fp[:-1])
    lin_dxp = xp[1:] - xp[:-1]
    b = log_dfp / lin_dxp
    a = fp[:-1] / np.exp(b * xp[:-1])

    out = np.zeros_like(fp, shape=x.shape)  # Shape of x, object style of fp

    for i in range(len(log_dfp)):
        slice = (xp[i] <= x) & (x <= xp[i + 1])
        out[slice] = a[i] * np.exp(b[i] * x[slice])

    sliceL, sliceR = x < xp[0], x > xp[-1]
    if bounded is True:
        out[sliceL] = fp[0]
        out[sliceR] = fp[-1]
    else:
        out[sliceL] = a[0] * np.exp(b[0] * x[sliceL])
        out[sliceR] = a[-1] * np.exp(b[-1] * x[sliceR])

    return out


def moving_average(x: Hint.nums, w: int = None):
    """
    Compute the moving average of an array.

    Args:
        x: 1D array which should have the moving average applied.
        w: The size of the moving average window. Optional, defaults to 2.

    Returns:
        An array of size n-(w-1) (when given an input array x of size n) and a
            moving average window size of w.

    """
    # Recast as necessary
    x = cast2numpy(x)
    w = 2 if w is None else int(2)

    # https://stackoverflow.com/questions/14313510/how-to-calculate-rolling-moving-average-using-python-numpy-scipy
    return np.convolve(x, np.ones(w), 'valid') / w


def point_diff(y: Hint.num, x: Hint.num = None) -> np.ndarray:
    """
    Given coordinate arrays describing discrete points, find the gradient at
    (and not between!) the points.

    Args:
        y: Y-values of the discrete points.
        x: X-values of the discrete points. Optional, assumes the form of
            range(len(y)) if not specified.

    Returns:
        The averaged differentials of the array, mapped back to the same shape
        as (,and the original location prescribed by) the inputs.

    """
    # Recast as necessary
    y = cast2numpy(y)
    x = np.arange(len(y)) if x is None else cast2numpy(x)

    assert x.shape == y.shape, "Expected homogeneity of array shapes"
    assert x.ndim == y.ndim == 1, f"Unsupported array dimensions, check for 1D"

    # Differentiation to give the gradients between points described in arrays
    dydx_mid = np.diff(y) / np.diff(x)

    # Gradient at the points is described by averaging the prior result
    dydx_pts = np.zeros(x.shape)
    # dydx_pts[0], dydx_pts[-1] = dydx_mid[0], dydx_mid[-1]  # copy ends
    dydx_pts[1:-1] = moving_average(dydx_mid, 2)
    dydx_pts[0], dydx_pts[-1] = dydx_pts[1], dydx_pts[-2]  # copy ends

    return dydx_pts


def point_curvature(x: Hint.num, y: Hint.num) -> np.ndarray:
    """
    Given arrays describing points in 2D, determine the signed curvature at each
    point.

    Args:
        x: 1D array of coordinates, the abscissa; x-coordinate.
        y: 1D array of coordinates, the ordinates; y-coordinate.

    Returns:
        An array describing the signed curvature at each point.

    """
    # Recast as array
    x = cast2numpy(x)
    y = cast2numpy(y)

    # Requisite differentations
    xprime = point_diff(x)
    xpprime = point_diff(xprime)
    yprime = point_diff(y)
    ypprime = point_diff(yprime)

    # Compute
    numerator = xprime * ypprime - yprime * xpprime
    denominator = (xprime ** 2 + yprime ** 2) ** (3 / 2)
    signed_curvature = numerator / denominator

    return signed_curvature



from carpy.utility import Quantity

__all__ = ["PhotovoltaicPanel", "Battery", "PSU"]


class PhotovoltaicPanel(object):

    def __init__(self):
        self._P_out = Quantity(200, "W")
        return

    def __repr__(self):
        return "PV"

    def P_out(self, P_in=None, **kwargs):
        return self._P_out


class PSU(object):
    _P_input_type = "electrical"

    def __init__(self, eta):
        self._eta = eta
        return

    def __repr__(self):
        return "PSU"

    def P_out(self, P_in=None, **kwargs):
        return P_in * self._eta


class Battery(object):
    _P_input_type = "electrical"

    def __init__(self, specific_energy=None):
        specific_energy = 80 if specific_energy is None else specific_energy
        self.specific_energy = Quantity(specific_energy, "W h kg^{-1}")
        return

    def __repr__(self):
        return "Battery"


"""Common methods for propulsion components."""

__all__ = ["EnergyForm"]
__author__ = "Yaseen Reza"


class Tag(object):

    def __init__(self):
        tagname = f"{type(self).__name__}"
        raise RuntimeError(f"Tag '{tagname}' is not meant to be instantiated")


# class Tags(object):
#     """
#     A class of tags to classify the type of value being conveyed. The tags in
#     this class exist as a means of unifying definitions of what a component of
#     the network understands and perceives as a valid input.
#
#     The point of a tag is that a number may represent a quantity like power, but
#     not all types of power are immediately interchangeable in a system. For
#     example:
#
#     -   A turbine may expect fluid (hydraulic) power as an input - and to
#         ensure consistency of the code's operation, a turbine must not be able
#         to accept electrical power as a valid input despite sharing the same
#         unit dimensions.
#     -   On the other hand, a method that multiplicatively applies efficiency to
#         the idealised power output of a component is fine to accept mechanical,
#         electrical, or thermal efficiencies - but must not accept process
#         efficiencies because the process of computation is more complicated.
#
#     """
#
#     @staticmethod
#     def value(tagged_value: tuple, expected_tag):
#         """
#         Extract the value from a tagged value, should it meet conditions.
#
#         Args:
#             tagged_value: Tuple in the form (value, tag).
#             expected_tag: The tag expected of the value
#
#         Returns:
#             The value, provided the tags match.
#
#         """
#
#         value, tag = tagged_value
#
#         if tag is expected_tag:
#             return value
#         else:
#             raise ValueError(f"{tagged_value} did not meet {expected_tag=}")
#
#     class Power(Tag):
#         """Classification: object represents a type of power."""
#         unit_equivalent = "W"
#
#         class Mechanical(Tag):
#             """
#             Classification: object is of mechanical power type.
#
#             Accounts for power (torque) transmitted in shafts and levers.
#             """
#             unit_equivalent = "W"
#
#         class Electrical(Tag):
#             """
#             Classification: object is of electrical power type.
#
#             Accounts for power transmitted in or used by electrical machines.
#             """
#             unit_equivalent = "W"
#
#         class Hydraulic(Tag):
#             """
#             Classification: object is of hydraulic power type.
#
#             Accounts for power transmitted through movement of fluid.
#             """
#             unit_equivalent = "W"
#
#         class Thermal(Tag):
#             """
#             Classification: object is of thermal power type.
#
#             Accounts for heating power and thermal losses.
#             """
#             unit_equivalent = "W"
#
#     class Efficiency(Tag):
#         """Classification: object represents an efficiency."""
#         unit_equivalent = ""
#
#         class Work(Tag):
#             """
#             Classification: Work-related efficiency.
#
#             Accounts for losses in power during transmission or transformation.
#             """
#             unit_equivalent = ""
#
#             class Mechanical(Tag):
#                 """
#                 Classification: object is a mechanical efficiency.
#
#                 Accounts for power loss, for example, in bearings/transmissions.
#                 """
#                 unit_equivalent = ""
#
#             class Electrical(Tag):
#                 """
#                 Classification: object is an electrical efficiency.
#
#                 Accounts for power loss, for example, in wiring/transformers.
#                 """
#                 unit_equivalent = ""
#
#             class Chemical(Tag):
#                 """
#                 Classification: object is a chemical efficiency.
#
#                 Accounts for power loss, for example, in fuel energy release.
#                 """
#                 unit_equivalent = ""
#
#         class Process(Tag):
#             """Classification: Thermodynamic process efficiency."""
#             unit_equivalent = ""
#
#             class Polytropic(Tag):
#                 """
#                 Classification: Process efficiency.
#
#                 Accounts for power loss through small-stage efficiency.
#                 """
#                 unit_equivalent = ""
#
#             class Isentropic(Tag):
#                 """
#                 Classification: Isentropic process efficiency.
#
#                 Accounts for power loss in isentropic processes
#                 """
#                 unit_equivalent = ""
#
#             # class Diabatic(object):
#             #     """
#             #     Classification: Diabatic process efficiency.
#             #
#             #     Accounts for power loss, for example, in diabatic ducts.
#             #     """
#             #     unit_equivalent = ""
#
#     class Conditions(Tag):
#         """Classifications of types of conditions applicable to fluid flow."""
#         unit_equivalent = NotImplemented
#
#         class State(Tag):
#             """Classification: object represents static conditions of fluid."""
#             unit_equivalent = NotImplemented
#
#         class MassFlowrate(Tag):
#             """Classification: object represents fluid mass flowrate."""
#             unit_equivalent = "kg s^{-1}"
#
#         class VolumetricFlowrate(Tag):
#             """Classification: object represents fluid volume flowrate."""
#             unit_equivalent = "m^{3} s^{-1}"
#
#     class RatioStateVars(Tag):
#         """Classification: Dimensionless ratio of thermodynamic state vars."""
#
#         class Pressure(Tag):
#             """Classification: object represents pressure ratio."""
#             unit_equivalent = ""
#
#         class Temperature(Tag):
#             """Classification: object represents temperature ratio."""
#             unit_equivalent = ""
#
#         class AdiabaticIndex(Tag):
#             """Classification: object represents specific heat ratio."""
#             unit_equivalent = ""


class EnergyForm(object):
    """
    References:
        -   Forms of energy, online, `eia.gov/energyexplained/what-is-energy/forms-of-energy.php`
            (Accessed 05/06/2023).
    """

    class Potential(Tag):
        """Stored energy and the energy of position."""

    # Energy in bonds
    Chemical = type("Chemical", (Potential,), dict())
    # Energy in tension
    Mechanical = type("Mechanical", (Potential,), dict())
    # Energy in nuclei
    Nuclear = type("Nuclear", (Potential,), dict())
    # Energy in work against gravity
    Gravitational = type("Gravitational", (Potential,), dict())

    class Kinetic(Tag):
        """Energy due to motion of waves, particles, substances, and objects."""

    # Energy in electromagnetic radiation
    Radiant = type("Radiant", (Kinetic,), dict())
    # Energy in movement of atoms and molecules
    Thermal = type("Thermal", (Kinetic,), dict())
    # Energy in object inertia
    Motion = type("Motion", (Kinetic,), dict())
    # Energy in compression/rarefaction waves in substances
    Sound = type("Sound", (Kinetic,), dict())
    # Energy in movement of charge
    Electrical = type("Radiant", (Kinetic,), dict())



"""Tests for atmosphere models in carpy."""
import unittest

import numpy as np

from carpy.environment import ISA1975


class ISA1975Quantities(unittest.TestCase):
    """Test the ISA1975 atmosphere model."""

    def test_TPD(self):
        """Test temperature, pressure, and density functions."""
        isa = ISA1975()

        T0 = isa.T(altitude=0)
        p0 = isa.p(altitude=0)
        rho0 = isa.rho(altitude=0)

        # Test units
        self.assertTrue(T0.units.si_equivalent == "K")
        self.assertTrue(p0.units.si_equivalent == "kg m^{-1} s^{-2}")
        self.assertTrue(rho0.units.si_equivalent == "kg m^{-3}")

        # Test sea-level values
        self.assertTrue(T0 == 288.15)
        self.assertTrue(p0 == 101325)
        self.assertTrue(round(rho0, 3) == 1.225)

        # Test geopotential values
        self.assertTrue(round(isa.p(11e3)) == 22632)
        self.assertTrue(round(isa.p(20e3)) == 5475)
        self.assertTrue(round(isa.p(32e3)) == 868)
        self.assertTrue(round(isa.p(47e3)) == 111)
        self.assertTrue(round(isa.p(51e3)) == 67)
        self.assertTrue(round(isa.p(71e3)) == 4)

        # Test geometric values
        self.assertTrue(round(isa.p(11e3 + 19, geometric=True)) == 22632)
        self.assertTrue(round(isa.p(20e3 + 63, geometric=True)) == 5475)
        self.assertTrue(round(isa.p(32e3 + 162, geometric=True)) == 868)
        self.assertTrue(round(isa.p(47e3 + 350, geometric=True)) == 111)
        self.assertTrue(round(isa.p(51e3 + 413, geometric=True)) == 67)
        self.assertTrue(round(isa.p(71e3 + 802, geometric=True)) == 4)

        # Test that functions can handle arrays for pressure (geopotential)
        goldx_0 = np.array([11e3, 20e3, 32e3, 47e3, 51e3, 71e3])
        goldx_1 = np.array([11019, 20063, 32162, 47350, 51413, 71802])

        goldy_p = np.array([22632, 5475, 868, 111, 67, 4])
        self.assertTrue(np.isclose(isa.p(goldx_0).x, goldy_p, atol=1e-1).all())

        # Test that functions can handle arrays for pressure (geometric)
        self.assertTrue(np.isclose(
            isa.p(goldx_1, geometric=True).x,
            goldy_p,
            atol=1e-1
        ).all())

        print()

        return

    def test_airspeeds(self):
        """Test conversions between airspeed types."""
        isa = ISA1975()

        goldspeeds = {
            "CAS": np.array([124.20454408, 377.49844212]),
            "EAS": np.array([100., 200.]),
            "TAS": np.array([415.60659989, 831.21319978]),
            "Mach": np.array([1.40418125, 2.8083625])
        }

        for i, (k, v) in enumerate(goldspeeds.items()):
            # 70,000 [ft] == 21_336 [m]
            speeds = isa.airspeeds(altitude=21_336, geometric=False, **{k: v})
            cas, eas, tas, mach = speeds

            self.assertTrue(np.isclose(goldspeeds["CAS"], cas, atol=1e-1).all())
            self.assertTrue(np.isclose(goldspeeds["EAS"], eas, atol=1e-1).all())
            self.assertTrue(np.isclose(goldspeeds["TAS"], tas, atol=1e-1).all())
            self.assertTrue(
                np.isclose(goldspeeds["Mach"], mach, atol=1e-1).all())



"""A module of miscellaneous library utilities."""
import functools
import inspect
import os
import typing

import numpy as np
import pandas as pd
import yaml

__all__ = []
__author__ = "Yaseen Reza"


# ============================================================================ #
# Data types and casting
# ---------------------------------------------------------------------------- #
class Hint(object):
    """A static class of common argument typehints."""
    iter = typing.Union[tuple, list, np.ndarray]
    num = typing.Union[int, float, np.integer, np.inexact]
    nums = typing.Union[iter, num]
    func = typing.Union[typing.Callable]
    any = typing.Union[typing.Any]
    set = typing.Union[set, frozenset]


def cast2numpy(scalar_or_vector: Hint.any, /, dtype=None) -> np.ndarray:
    """
    Args:
        scalar_or_vector: Scalar or vector numerical argument to be recast.
        dtype: Datatype to cast as. Optional, defaults to None.

    Returns:
        The input argument recast as an array (if not already array like).

    Raises:
        ValueError: Couldn't determine a way to cast input as an array.

    """
    type_transforms = {
        (int, float, complex): lambda x: np.array([x]),
        (np.integer, np.inexact): lambda x: np.array([x]),
        (tuple, list, range): lambda x: np.array(x),
        (set, frozenset): lambda x: np.array([xi for xi in x]),
        (dict,): lambda x: {k: cast2numpy(v) for k, v in x.items()},
        (map, filter): lambda x: np.array(list(x)),
        (np.ndarray,): lambda x: x if x.shape != () else np.array([x]),
        (pd.Series,): lambda x: np.array(x)
    }
    # Look for and try to apply a transform
    for types, transformer in type_transforms.items():

        if not isinstance(scalar_or_vector, types):
            continue  # Try a different transformer

        # No dtype specified, do not attempt to apply any casting rules
        if dtype is None:
            return transformer(scalar_or_vector)
        else:
            return transformer(scalar_or_vector).astype(dtype=dtype)

    # We tried all that we could
    errormsg = f"Couldn't cast '{type(scalar_or_vector)}' to numpy array"
    raise ValueError(errormsg)


def isNone(*args) -> tuple:
    """True or False for every arg that is None."""
    results = tuple(map(lambda x: x is None, args))
    return results if len(args) > 1 else results[0]


def collapse_array(scalar_or_vector):
    """
    Given a vector (or scalar), collapse any redundant nesting.

    Args:
        scalar_or_vector: An iterable to collapse.

    Returns:
        Lower dimensional representation of the input.

    """
    # If the "scalar_or_vector" is not iterable, simply return it
    if not isinstance(scalar_or_vector, Hint.iter.__args__):
        return scalar_or_vector

    # Else it is iterable. Do we lose anything for indexing?
    if len(scalar_or_vector) == 1:
        return collapse_array(scalar_or_vector[0])  # Recursively collapse

    # Yes, information is lost with further indexing. Return as is.
    else:
        return scalar_or_vector


__all__ += [Hint.__name__, cast2numpy.__name__, isNone.__name__,
            collapse_array.__name__]


# ============================================================================ #
# Pathing and files
# ---------------------------------------------------------------------------- #
class GetPath(object):
    """A static class of methods providing relevant file paths."""
    _library_path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), os.pardir))

    @classmethod
    def library(cls) -> str:
        """Return the absolute OS path of the library's installation."""
        return cls._library_path

    @staticmethod
    def localpackage(*optional_paths) -> str:
        """
        Return the absolute path of the caller file's local package.

        Args:
            *optional_paths: Optional extra paths to follow.

        Returns:
            Returns 'os.path.join(localpackage_path, *paths)'.

        """
        local_pkg_path = os.path.dirname(inspect.stack()[1].filename)
        extended_path = os.path.join(local_pkg_path, *optional_paths)
        return extended_path

    @staticmethod
    def github() -> str:
        """Return the URL to carpy's source."""
        return "https://github.com/yaseen157/carpy"


class LoadData(object):
    """A static class of methods to load various file types"""

    @staticmethod
    def yaml(filepath) -> dict:
        """Provided with a filepath, return a dictionary of discovered data."""
        with open(filepath, "r") as stream:
            try:
                dict2return = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
        return dict2return

    @staticmethod
    def excel(filepath, **kwargs) -> pd.DataFrame:
        """Provided with a filepath, return a DataFrame obj (kwargs apply)."""
        overwritable_defaults = {"sheet_name": None, "na_filter": False}
        updatedkwargs = {**overwritable_defaults, **kwargs}
        dataframes = pd.read_excel(io=filepath, **updatedkwargs)
        return dataframes


__all__ += [GetPath.__name__, LoadData.__name__]


# ============================================================================ #
# Decoration
# ---------------------------------------------------------------------------- #
def call_count(func):
    """Decorator to count the number of times a function is called."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        """Function wrapper that counts the number of function calls."""
        wrapper.call_count += 1
        return func(*args, **kwargs)

    wrapper.call_count = 0
    return wrapper


def call_depth(func):
    """
    Decorator to track the recursive depth of a function call.

    Args:
        func: Function to track the recursive entries of.

    Returns:
        Wrapped function.

    Examples:

        # Recursively multiply by two by as many times as desired
        >>> @call_depth
        ... def multiply_by_2(x, n_times=1):
        ...     # If maximum call depth is reached, ascend recursive stack
        ...     if multiply_by_2.call_depth == n_times:
        ...         return x
        ...     return 2 * multiply_by_2(x, n_times)

        >>> print(multiply_by_2(3, n_times=3))
        24

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        """Function wrapper that tracks the depth of function recursion."""
        wrapper.call_depth += 1
        result = func(*args, **kwargs)
        wrapper.call_depth -= 1
        return result

    wrapper.call_depth = -1
    return wrapper


__all__ += [call_count.__name__, call_depth.__name__]



"""A module of various methods used to estimate wing aerodynamic performance."""
import numpy as np
from scipy.integrate import simpson, trapezoid

from carpy.aerodynamics.aerofoil import ThinAerofoil
from carpy.environment import ISA1975
from carpy.structures import DiscreteIndex
from carpy.utility import Hint, Quantity, cast2numpy, moving_average

__all__ = ["PLLT", "HorseshoeVortex", "CDfGudmundsson"]  # Cantilever1DStatic"]
__author__ = "Yaseen Reza"


# ============================================================================ #
# Support Functions and Classes
# ---------------------------------------------------------------------------- #


def designate_sections(sections: DiscreteIndex, mirror: bool = None,
                       N: int = None) -> list:
    """
    Given a WingSections object, distribute sections over N-elements.

    Args:
        sections (WingSections): WingSections object, describing wing geometry.
        mirror: If set to True, the given sections should be mirrored to
            produce a symmetrical wing. Optional, defaults to True.
        N: The number of sections to return. Optional, defaults to 40.

    Returns:
        list: A list containing discretised WingSection elements.

    """
    # Recast as necessary
    mirror = True if mirror is None else mirror
    N = 40 if N is None else int(N)

    # Linearly spaced sections in defined span
    if mirror is True:
        if N % 2 == 0:  # Even number of horseshoes, don't define @y=0
            sections2rtn = sections[::N][1::2]
            sections2rtn = sections2rtn[::-1] + sections2rtn
        else:  # Odd number of horseshoes, definition @y=0 is required
            sections2rtn = sections[::int(np.ceil(N / 2))]  # ceil includes y=0
            sections2rtn = sections2rtn[:0:-1] + sections2rtn
    else:
        sections2rtn = sections[::N]

    return sections2rtn


def vfil(xyzA: Hint.nums, xyzB: Hint.nums, xyzC: Hint.nums):
    """
    Compute the influence of vortex line AB at control point C, return zero for
    small denominators (to exclude self-influence).

    Args:
        xyzA: Vector describing position of reference point A in 3D space.
        xyzB: Vector describing position of reference point B in 3D space.
        xyzC: Vector describing position of control point C in 3D space.

    Returns:
        Influence I of vortex line AB at control point C.

    """
    # Recast as necessary
    xyzA, xyzB, xyzC = cast2numpy([xyzA, xyzB, xyzC])

    small = 1e-12
    ab = xyzB - xyzA
    ac = xyzC - xyzA
    bc = xyzC - xyzB
    cross = np.cross(ac, bc)
    den = 4 * np.pi * (cross ** 2).sum()
    abs_ac = np.sqrt((ac ** 2).sum())
    abs_bc = np.sqrt((bc ** 2).sum())
    num = (ab * (ac / abs_ac - bc / abs_bc)).sum()

    if den <= small:
        influence = np.zeros(3)
    else:
        influence = num / den * cross

    return influence


class WingSolution(object):
    """
    Template object, of which solvers should aim to produce all attributes of.
    """
    _AR: float
    _CL: float
    _CLalpha: float
    _Sref: Quantity
    _b: Quantity
    _e: float
    _delta: float
    _tau: float

    def __str__(self):
        params2print = ["AR", "CL", "CLalpha", "Sref", "b", "e", "delta", "tau"]

        return_string = f"{type(self).__name__}\n"
        for param in params2print:
            if hasattr(self, f"_{param}"):
                return_string += f">\t{param:<8} = {getattr(self, param)}\n"

        return return_string

    @property
    def AR(self) -> float:
        """Wing aspect ratio, AR."""
        return self._AR

    @property
    def CL(self) -> float:
        """Wing lift coefficient, CL."""
        return self._CL

    @property
    def CLalpha(self) -> float:
        """Finite wing lift-curve slope, CLalpha."""
        return self._CLalpha

    @property
    def Sref(self) -> Quantity:
        """Wing reference planform area, Sref."""
        return self._Sref

    @property
    def b(self) -> Quantity:
        """Tip-to-tip span of the wing, b."""
        return self._b

    @property
    def e(self) -> float:
        """Planform span efficiency factor, e."""
        return self._e

    @property
    def delta(self) -> float:
        """Elliptical lift distribution deviation factor, delta."""
        return self._delta

    @property
    def tau(self) -> float:
        """Finite wing lift-curve slope deviation factor, tau."""
        return self._tau


# ============================================================================ #
# Public classes
# ---------------------------------------------------------------------------- #

class PLLT(WingSolution):
    """
    Prandtl's lifting line (a.k.a Lanchester-Prandtl wing) theory.

    For use in incompressible, inviscid, steady flow regimes with moderate to
    high aspect ratio, unswept wings.
    """

    def __init__(self, sections: DiscreteIndex, span: Hint.num, alpha: Hint.num,
                 N: int = None):
        # Recast as necessary
        N = 50 if N is None else int(N)

        # Create a cosine-spaced distribution of the given extent of sections
        section_max, section_min = max(sections), min(sections)

        # Skip first element to prevent the generation of singular matrices
        theta0 = np.linspace(0, np.pi, N + 2)[1:-1]

        # Station parameters (geometric)
        y = np.cos(theta0) * (span / 2)  # ... Span position
        Nsections = sections[
            np.interp(abs(y), [0, span / 2], [section_min, section_max])
        ]
        chord = [Nsection.chord.x for Nsection in Nsections]
        # Station parameters (aerodynamic)
        alpha_zl, Clalpha = [], []
        for i, Nsection in enumerate(Nsections):
            solution = ThinAerofoil(
                aerofoil=Nsection.aerofoil,
                alpha=alpha + Nsection.twist  # Effective AoA
            )
            alpha_zl.append(solution.alpha_zl - Nsection.twist)  # Effective ZL
            Clalpha.append(solution.Clalpha)

        chord, alpha_zl, Clalpha = map(cast2numpy, (chord, alpha_zl, Clalpha))

        # Solve matrices for Fourier series coefficients
        matA = np.zeros((N, N))
        for j in (n := np.arange(N)):
            term1 = 4 * span / Clalpha[j] / chord[j]
            term2 = (n + 1) / np.sin(theta0[j])
            matA[j] += np.sin((n + 1) * theta0[j]) * (term1 + term2)
        matB = (alpha - alpha_zl)[:, None]
        matX = np.linalg.solve(matA, matB)[:, 0]  # <-- This is the slow step!!!

        # Just some requisite quick maths :)
        wingarea = -simpson(y=chord, x=y)
        a0 = 2 * np.pi  # Ideal lift-curve slope
        n = (np.arange(1, len(matX) + 1))
        delta = (n * (matX / matX[0]) ** 2)[1:].sum()
        # n increases row-wise, theta0 is periodic column-wise
        # Take -ve of simpson integral because x=y moves from y=+b/2 to y=-b/2
        integrals = -simpson(y=chord * np.sin(n[:, None] * theta0), x=y)
        tau = span / (2 * wingarea) * (n * matX * integrals)[1:].sum()

        self._AR = span ** 2 / wingarea
        self._CL = np.pi * self.AR * matX[0]
        self._CLalpha = a0 / (1 + a0 / (np.pi * self.AR) * (1 + tau))
        self._Sref = Quantity(wingarea, "m^{2}")
        self._b = Quantity(span, "m")
        self._e = 1 / (1 + delta)
        self._delta = delta
        self._tau = tau

        return


class HorseshoeVortex(WingSolution):
    """
    Horseshoe Vortex Method over N-elements.

    For use in incompressible, inviscid, steady flow regimes.

    Notes:
        Currently assumes all lifting sections have a lift-slope of 2 pi.

    """

    def __init__(self, sections: DiscreteIndex, span: Hint.num, alpha: Hint.num,
                 N: int = None, mirror: bool = None):
        """
        Args:
            sections: Wing sections object, defining 3D geometry.
            span: Tip-to-tip span of wing (if mirrored), else root-to-tip span.
            alpha: Angle of attack at the wing root (station zero).
            N: The number of horseshoe vortices. Optional, defaults to 40.
            mirror: Whether or not to treat the wing sections object as
                symmetrical around the smallest station index. Optional,
                defaults to True (mirrored wing).

        Notes:
            The reference coordinate system used in this code adheres to the
            NorthEastDown (NED) standard aircraft principle axes representation.

        """
        # Recast as necessary
        N = 40 if N is None else int(N)
        mirror = True if mirror is None else mirror

        if mirror is False:
            raise NotImplementedError("Sorry, this method isn't ready yet.")

        # Problem setup
        large = 1e6
        Q = np.array([-np.cos(alpha), 0.0, -np.sin(alpha)])  # Freestream orient
        scalefactor = np.sqrt(N / (N + 1))  # Scaling fix for N = 1 case
        # Linearly spaced sections in defined span
        Nsections = designate_sections(sections=sections, mirror=mirror, N=N)
        self._Nsections = Nsections
        # bprime = combined span of all horseshoe elements
        bprime = span * scalefactor
        # cprime = longitudinal dist. between control point and horseshoe front
        chord = np.array([section.chord.x for section in Nsections])
        cprime = np.array([[-0.5 * cw / scalefactor, 0, 0] for cw in chord])

        # Locate vortex segments (a,b), control vectors (c), and normal vectors
        va, vb, vc, n = np.zeros((4, N, 3))  # Initial assignment to 3D zeros.
        # ... distribute vortex locators (va, vb) in the y-direction
        if mirror is True:
            ys = np.linspace(-bprime / 2, bprime / 2, N + 1)
        else:
            ys = np.linspace(0, bprime, N + 1)
        va[:, 1] = ys[:-1]
        vb[:, 1] = ys[1:]
        # ... prescribe normal vectors as if wing is a flat plate
        n[:, 2] = -1.0

        # Locate semi-infinite vortices
        vinf = np.zeros((N, 3))
        vinf[:, 0] = -large

        # Define horseshoe geometry in the y >= 0, +ve direction
        dy = bprime / N
        for i in range(N):

            solution = ThinAerofoil(
                aerofoil=Nsections[i].aerofoil,
                alpha=alpha + Nsections[i].twist  # Effective AoA
            )
            alpha_zl = solution.alpha_zl - Nsections[i].twist  # Effective ZL

            # Port-side vortices (any mirrored elements, if they exist)
            if va[i, 1] < 0 and vb[i, 1] <= 0:
                continue  # skip, we'll populate this later

            # Centreline vortex (vortex symmetrical about the centreline)
            elif va[i, 1] < 0:
                # Create rotation matrix for aerodynamic + geometric twist
                # section.alpha_zl automatically combines aero + geo twist...
                cos_theta = np.cos(-alpha_zl)
                sin_theta = np.sin(-alpha_zl)
                rot_twist = np.array([
                    [cos_theta, 0, sin_theta],
                    [0, 1, 0],  # About the Y-axis
                    [-sin_theta, 0, cos_theta]
                ])
                del cos_theta, sin_theta

                # Apply rotation to normal vector and control point locator
                n[i] = rot_twist @ n[i]
                vc[i] = (va[i] + vb[i]) / 2 + (rot_twist @ cprime[i])
                vinf[i] = rot_twist @ vinf[i]  # Twist semi-infinite vortices
                # Dihedral and sweep are safely ignored at the centreline

            # Starboard vortices
            else:
                # Sweep vortex locators (Move x-coordinates)
                # ... this assumes that the vortex locators (which are more
                # ... akin to centre of pressure locators) sweep by the same
                # ... amount that the leading edge does. Not true in real life!
                tan_sweep = np.tan(Nsections[i].sweep)
                va[i + 1:, 0] -= dy * tan_sweep  # Outboard sections
                vb[i + 1:, 0] -= dy * tan_sweep
                va[i, 0] -= (dy / 2) * tan_sweep  # Current section
                vb[i, 0] -= (dy / 2) * tan_sweep
                del tan_sweep

                # Apply dihedral to vortex locators (Move z-coordinates)
                tan_dihedral = np.tan(Nsections[i].dihedral)
                va[i + 1:, 2] -= dy * tan_dihedral  # Outboard sections
                vb[i + 1:, 2] -= dy * tan_dihedral
                va[i, 2] -= (dy / 2) * tan_dihedral  # Current section
                vb[i, 2] -= (dy / 2) * tan_dihedral
                del tan_dihedral

                # Create rotation matrix for aerodynamic + geometric twist
                # section.alpha_zl automatically combines aero + geo twist...
                cos_theta = np.cos(-alpha_zl)
                sin_theta = np.sin(-alpha_zl)
                rot_twist = np.array([
                    [cos_theta, 0, sin_theta],
                    [0, 1, 0],  # About the Y-axis
                    [-sin_theta, 0, cos_theta]
                ])
                del cos_theta, sin_theta

                # Create rotation matrix for dihedral
                # Take negatives, because matrix rotation assumes +ve==clockwise
                cos_gamma = np.cos(-Nsections[i].dihedral)
                sin_gamma = np.sin(-Nsections[i].dihedral)
                rot_dihedral = np.array([
                    [1, 0, 0],  # About the X-axis
                    [0, cos_gamma, -sin_gamma],
                    [0, sin_gamma, cos_gamma]
                ])
                del cos_gamma, sin_gamma

                # Apply rotations to normal vector and control point locator
                n[i] = rot_dihedral @ (rot_twist @ n[i])
                vc[i] = (va[i] + vb[i]) / 2 + (rot_twist @ cprime[i])
                # apparently, the below step is controversial???
                vinf[i] = rot_twist @ vinf[i]  # Twist semi-infinite vortices

        # Now if necessary (wing mirrored), copy starboard geometry to port side
        else:
            if mirror is True:
                mirror_idx = int(N / 2)
                # For consistency, va and vb must swap in the mirror image
                va[:mirror_idx] = vb[-mirror_idx:][::-1] * np.array([1, -1, 1])
                vb[:mirror_idx] = va[-mirror_idx:][::-1] * np.array([1, -1, 1])
                # ", but mirroring vc
                vc[:mirror_idx] = vc[-mirror_idx:][::-1] * np.array([1, -1, 1])
                # Simple mirror for normal vectors
                n[:mirror_idx] = n[-mirror_idx:][::-1]

        # Set matrix and RHS elements, solve for circulations (Gamma)
        A = np.zeros((N, N))
        rhs = np.zeros((N, 1))

        for i in range(N):
            for j in range(N):
                # Influence contributions; 'LARGE -> va -> vb -> LARGE' on vc
                infl = vfil(va[j], vb[j], vc[i])
                infl += vfil(va[j] + vinf[j], va[j], vc[i])
                infl += vfil(vb[j], vb[j] + vinf[j], vc[i])
                A[i, j] = (infl * n).sum()
            rhs[i] = -(Q * n).sum()

        # Sum of circulation in z-direction should match freestream, given gamma
        gamma = np.linalg.solve(A, rhs)

        # Forces at centres of bound vortices (parallel with leading edge)
        Fx, Fy, Fz = np.zeros((3, N))
        self._xyz_cp = 0.5 * (va + vb)
        for i in range(N):
            # Local velocity vector on each load element i
            u = np.copy(Q)
            for j in range(N):
                u += vfil(va[j], vb[j], self._xyz_cp[i]) * gamma[j]
                u += vfil(va[j] + vinf[j], va[j], self._xyz_cp[i]) * gamma[j]
                u += vfil(vb[j], vb[j] + vinf[j], self._xyz_cp[i]) * gamma[j]
            # u cross s gives direction of action of the force from circulation
            s = vb[i] - va[i]
            Fx[i], Fy[i], Fz[i] = np.cross(u, s) * gamma[i]

        # Resolve these forces into perpendicular and parallel to freestream
        elems = 1000
        area_chords = [qty.x for qty in sections[::elems].chord]
        wingarea = trapezoid(area_chords, dx=span / elems)
        halfS = wingarea / 2
        self._Cl = (Fx * np.sin(alpha) - Fz * np.cos(alpha)) / halfS
        self._Cdi = (-Fx * np.cos(alpha) - Fz * np.sin(alpha)) / halfS

        self._AR = span ** 2 / wingarea
        self._CL = self._Cl.sum()
        self._CLalpha = NotImplemented
        self._Sref = Quantity(wingarea, "m^{2}")
        self._b = Quantity(span, "m")
        self._e = self._CL ** 2 / np.pi / self._AR / self._Cdi.sum()
        self._delta = 1 / self._e - 1
        self._tau: float = NotImplemented

        return

    @property
    def Cl(self) -> np.ndarray:
        """
        An array describing the distribution of generated CL. This is from
        port to starboard in a full wingplane, or from an inboard section to the
        outboard sections.

        Returns:
            Section-wise distribution of CL components.

        """
        return self._Cl

    @property
    def xyz_cp(self) -> np.ndarray:
        """
        An array of points, describing the distribution of wing section centres
        of pressures in physical space.

        Returns:
            Array of (N,3) points, describing xyz coordinates of sectional
            centres of pressure.

        """
        return self._xyz_cp


class Cantilever1DStatic(WingSolution):

    def __init__(self, sections, spar, span: Hint.num, alpha: Hint.num,
                 lift: Hint.num, N: int = None, mirror: bool = None,
                 model: str = None):
        # Define compatability parameters
        supported_models = [HorseshoeVortex]
        supported_models = dict(zip(
            [x.__name__ for x in supported_models],
            supported_models
        ))

        # Recast as necessary
        kwargs = {
            "sections": sections, "spar": spar, "span": span, "alpha": alpha,
            "lift": lift, "N": 40 if N is None else int(N),
            "mirror": True if mirror is None else mirror,
            "model": model
        }
        if kwargs["model"] is None:  # Model unspecified
            kwargs["model"] = HorseshoeVortex
        elif supported_models.get(model) is None:  # Model unsupported
            errormsg = f"{model=} is unsupported, try one of {supported_models}"
            raise ValueError(errormsg)

        # Evaluate aerodynamic model
        # noinspection PyUnresolvedReferences
        soln_aero = kwargs["model"](
            **{  # Iterate over kwargs, and pass only those which appear in init
                k: v for (k, v) in kwargs.items()
                if k in kwargs["model"].__init__.__annotations__  # <-Typehints!
            }
        )

        # Find out if solution is mirrored
        ys = soln_aero.xyz_cp[:, 1]
        if kwargs["mirror"] is True:
            root_i = int(np.ceil(len(ys) / 2))
            ys = ys[root_i:]
        else:
            root_i = 0

        # Lift and moment component distribution
        sectionL = lift * (soln_aero.Cl[root_i:] / soln_aero.CL)
        sectionM = sectionL * ys

        # Flexural modulus of carbon fibre (?)
        E_rect = 60e9  # e9 == GPa
        # Subscript x as it's the effect of material distributed about x-axis
        t_wall = 1e-3
        b_spar = 50e-3
        h0_spar = 100e-3
        h1_spar = 20e-3

        def Ix_rect(y):
            """Variable cross-section, hollow rectangular spar."""
            y = np.abs(y)
            h_spar = np.interp(y, [0, 10], [h0_spar, h1_spar])
            Ixx_o = b_spar * h_spar ** 3 / 12
            Ixx_i = (b_spar - 2 * t_wall) * (h_spar - 2 * t_wall) ** 3 / 12
            return Ixx_o - Ixx_i

        EI = E_rect * Ix_rect(ys)

        # Distribution of angular increments with each section
        sectiontheta = sectionM / EI

        # Distibution of vertical displacement with each section
        sectionnu = np.cumsum(sectiontheta)

        plot_dy, = np.diff(ys[0:2])
        plot_y = np.zeros_like(sectionnu)
        plot_z = np.copy(sectionnu)

        for i in range(1, len(sectiontheta)):
            plot_y[i:] += (plot_dy ** 2 - sectiontheta[i] ** 2) ** 0.5

        from matplotlib import pyplot as plt

        fig, axs = plt.subplots(3, dpi=140)
        axs[0].plot((plot_y := np.linspace(0, ys[-1], len(ys))), plot_z)
        axs[0].set_aspect(1)
        axs[1].plot(plot_y, sectionL)
        axs[2].plot(plot_y, sectionM)

        axs[0].set_ylabel("z(y)")
        axs[1].set_ylabel("L(y)")
        axs[2].set_ylabel("M(y)")

        # plt.show()

        return


class CDfGudmundsson(object):

    def __init__(self, sections: DiscreteIndex, span: Hint.num,
                 altitude: Hint.num, TAS: Hint.num, geometric: bool = None,
                 atmosphere=None, N: int = None, mirror: bool = None):
        """

        Args:
            sections ():
            span ():
            altitude ():
            TAS ():
            geometric:
            atmosphere ():
            N ():
            mirror ():

        Notes:
            Section 15.3.3, Step-by-step calculation of skin friction drag
        """
        # Recast as necessary
        span = Quantity(span, "m")
        TAS = Quantity(TAS, "m s^{-1}")
        geometric = False if geometric is None else geometric
        atmosphere = ISA1975() if atmosphere is None else atmosphere
        N = 40 if N is None else int(N)
        mirror = True if mirror is None else mirror

        # Assume roughness of carefully applied matte paint
        from carpy.utility import constants as co
        kappa = co.MATERIAL.esg_roughness.mattepaint_careful.mean()

        # Linearly spaced sections in defined span (no need to mirror)
        Nsections = designate_sections(sections=sections, mirror=False, N=N)

        # Step 1) Find Viscosity of air
        mu_visc = atmosphere.mu_visc(altitude=altitude, geometric=geometric)

        # Step 2) Find Reynolds Number
        rho = atmosphere.rho(altitude=altitude, geometric=geometric)
        chords = Quantity([section.chord.x for section in Nsections], "m")
        Re = rho * TAS * chords / mu_visc

        # Step 3) Cutoff Reynolds number due to surface roughness effects
        Mach = TAS / atmosphere.c_sound(altitude=altitude, geometric=geometric)
        if Mach <= 0.7:
            Re_cutoff = 38.21 * (chords / kappa).x ** 1.053
        else:
            Re_cutoff = 44.62 * (chords / kappa).x ** 1.053 * Mach ** 1.16

        # noinspection PyArgumentList
        Re = np.vstack((Re, Re_cutoff)).min(axis=0)

        # Step 4) Compute skin friction coefficient for fully laminar/turbulent
        self._Cf_lam = 1.328 * Re ** -0.5
        self._Cf_turb = 0.455 * np.log10(Re) ** -2.58
        self._Cf_turb *= (1 + 0.144 * Mach ** 2) ** -0.65  # Compressible corr.

        # Step 5) Determine fictitious turbulent boundary layer origin point X0
        Xtr_C = 0.5  # Assume transition @X==0.5, avg of upper/lower transitions
        X0_C = 36.9 * Xtr_C ** 0.625 * Re ** -0.375

        # Step 6) Compute mixed laminar-turbulent flow skin friction coefficient
        Cf = 0.074 * Re ** -0.2 * (1 - (Xtr_C - X0_C)) ** 0.8

        # Find chords between stations Cf has been evaluated at
        b_eval = span.x / 2 if mirror is True else span.x
        ys = np.concatenate(
            ([0], moving_average(np.linspace(0, b_eval, N), w=2), [b_eval]))
        stations = np.interp(ys, [0, b_eval], [min(sections), max(sections)])
        midchords = np.array([sec.chord.x for sec in sections[stations]])
        # Find component areas of wing locally surrounding the defined sections
        Srefs = moving_average(midchords, w=2) * np.diff(ys)
        Swets = np.array([sec.aerofoil.perimeter for sec in Nsections]) * Srefs

        # Step 7) Compute skin friction drag coefficient
        Swet = Swets.sum()
        Sref = Srefs.sum()
        self._Cf = (Cf * Swets).sum() / Swet
        self._CDf = self._Cf * (Swet / Sref)

        return

    @property
    def CDf(self):
        """Wing skin friction drag coefficient, CDf."""
        return self._CDf



"""Tests for unit conversions methods."""
import unittest

import numpy as np

from carpy.utility import Quantity


class Quantities(unittest.TestCase):

    def test_instantiation(self):
        """Verify Quantity can be instantiated with many arguments."""
        testcases = (1, (1,), {1}, [1], np.array(1), np.array([1]))

        # No unit instantiation should just return an array, no questions asked
        for testcase in testcases:
            array_instance = Quantity(testcase)
            self.assertEqual(array_instance, 1)
            self.assertTrue(isinstance(array_instance, np.ndarray))
            self.assertTrue(isinstance(array_instance, Quantity))

        return

    def test_conversions(self):
        testcases = (
            # Temperature
            ((288.15, "K"), (288.15, "K")),
            ((288.15, "K"), (15.0, "degC")),
            ((288.15, "K"), (59.0, "degF")),
            # Power
            ((1, "nW"), (1e-9, "W")),
            ((1, "uW"), (1e-6, "W")),
            ((1, "mW"), (1e-3, "W")),
            ((1, "W"), (1, "W")),
            ((1, "kW"), (1e3, "W")),
            ((1, "MW"), (1e6, "W")),
            ((1, "GW"), (1e9, "W")),
            ((1, "hp"), (745.69987158, "W")),
            ((31, "dBm"), (1, "dBW")),
            ((0, "dBW"), (1, "W")),
            ((30, "dBm"), (1, "W")),
            # Pressure
            ((1, "bar"), (1e5, "Pa")),
            ((1, "bar"), (750.06157585, "mmHg")),
            ((1, "atm"), (29.92125558, "inHg")),
            # Homonuclear
            ((1, "mm^{2}"), (1e-6, "m^{2}"))

        )
        # Test conversions
        for ((arg_val, arg_unit), (gold_val, tgt_unit)) in testcases:
            self.assertTrue(
                expr=np.isclose(
                    Quantity(arg_val, arg_unit).to(tgt_unit),
                    gold_val
                ),
                msg=(
                    f"Failed to convert {arg_val} {arg_unit} to "
                    f"{gold_val} {tgt_unit}"
                )
            )
        return

    def test_operations(self):
        """Test mathematical and operational rigour of Quantity objects."""
        return


if __name__ == '__main__':
    unittest.main()



"""Means of constructing graphs (the networked mathematical kind)."""
import numpy as np

__all__ = ["Graphs"]
__author__ = "Yaseen Reza"


class Traverse(object):
    """Graph traversal methods."""

    @classmethod
    def depth_first(cls, graph: dict, source: object) -> set:
        """
        Perform depth-first-search (DFS) to find connected nodes in a graph.

        Args:
            graph: A dictionary of nodes (keys) and connected nodes (values).
            source: The node (key) from which to start searching the graph.

        Returns:
            set: Nodes discovered on graph traversal.

        Examples:

            >>> mygraph = Graph()
            >>> A = mygraph.new_node("A")
            >>> B = mygraph.new_node("B")
            >>> C = mygraph.new_node("C")
            >>> mygraph.new_link(A, B, directed=True)
            >>> mygraph.new_link(A, C, directed=True)
            >>> mygraph.new_link(B, C)

            >>> # The following nodes are traversed when searching from source
            >>> print(Traverse.depth_first(mygraph.link_map, B))
            {Node('C'), Node('B')}

        References:
            https://medium.com/geekculture/depth-first-search-dfs-algorithm-with-python-2809866cb358

        """
        # Initialise...
        visited = set()
        dfs_traversal = list()

        def dfs(node_source):
            """Helper function for recursive searching."""
            # Is the node we're visiting right now a new node?
            if node_source not in visited:
                dfs_traversal.append(node_source)
                visited.add(node_source)

                for neighbor_node in graph[node_source]:
                    dfs(neighbor_node)

        # Execute search
        dfs(node_source=source)

        return visited

    @classmethod
    def breadth_first(cls, graph: dict, source: object) -> set:
        """
        Perform breadth-first-search (BFS) to find connected nodes in a graph.

        Args:
            graph: A dictionary of nodes (keys) and connected nodes (values).
            source: The node (key) from which to start searching the graph.

        Returns:
            set: Nodes discovered on graph traversal.

        Examples:

            >>> mygraph = Graph()
            >>> A = mygraph.new_node("A")
            >>> B = mygraph.new_node("B")
            >>> C = mygraph.new_node("C")
            >>> mygraph.new_link(A, B, directed=True)
            >>> mygraph.new_link(A, C, directed=True)
            >>> mygraph.new_link(B, C)

            >>> # The following nodes are traversed when searching from source
            >>> print(Traverse.breadth_first(mygraph.link_map, B))
            {Node('C'), Node('B')}

        References:
            https://www.educative.io/answers/how-to-implement-a-breadth-first-search-in-python

        """
        # Initialise...
        visited = {source}  # Track visited nodes
        queue = [source]  # Spawn a queue

        while queue:
            node_source = queue.pop(0)

            # Visit all neighbours of this node
            for neighbour in graph[node_source]:
                if neighbour not in visited:
                    visited.add(neighbour)
                    queue.append(neighbour)

        return visited


class Graph(object):
    """Class for creating networks of vertices (nodes) and edges (links)."""

    def __init__(self):
        self._graph = dict()  # Nodes and their neighbours
        self._edges = dict()  # Nodes and their *visitable* neighbours
        return

    @property
    def node_map(self) -> dict:
        """
        Returns:
            A dictionary of every node in the graph, and all connected
                neighbours (including those that cannot be traversed to due to
                directed edges).

        """
        return self._graph

    @property
    def link_map(self) -> dict:
        """
        Returns:
            A dictionary of every node that can be traversed to from a given
                node in a graph, accounting for directed edges.

        """
        return self._edges

    class Node(object):
        """A class for creating vertex (or node) element of the graph."""

        def __init__(self, obj: object):
            """
            Args:
                obj: Object to assign to this graph element/vertex.
            """
            self._obj = obj
            return

        def __repr__(self):
            reprstring = f"{type(self).__name__}({self._obj.__repr__()})"
            return reprstring

        @property
        def obj(self):
            """The object assigned to this graph element."""
            return self._obj

    def new_node(self, obj: object) -> Node:
        """
        Spawn a new vertex in the graph.

        Args:
            obj: The user-object to be assigned to the node.

        Returns:
            The node object that was assigned to the graph. It's recommended
            that you pass this node object to the 'new_link' method..

        """
        new_node = self.Node(obj)
        self._graph[new_node] = []
        self._edges[new_node] = []
        return new_node

    def new_link(self, node0: Node, node1: Node, directed: bool = None) -> None:
        """
        Define a directed or undirected edge between two vertices of the graph.

        Args:
            node0: One node object present in the graph.
            node1: Another node object present in the graph.
            directed: Boolean flag. If true, information may travel from node0
                to node1, but *not* vice versa. Optional, defaults to False
                (undirected path).

        Returns:
            None.

        """
        # Recast as necessary
        directed = False if directed is None else directed

        # Verify that the nodes are members of this graph
        if node0 not in self._graph or node1 not in self._graph:
            errormsg = f"One of more given nodes were not members of this graph"
            raise ValueError(errormsg)

        # Add neighbours
        self._graph[node0].append(node1)
        self._graph[node1].append(node0)

        # Add visitable paths
        self._edges[node0].append(node1)
        # Add the graph edge from node1 to node0 if the edge is not directed
        if directed is False:
            self._edges[node1].append(node0)

        return None

    @property
    def mat_adjacency(self) -> np.ndarray:
        """
        Adjacency matrix of vertices in the graph.

        Returns:
            A numpy array in which enumerating rows is equivalent to enumerating
                nodes from which traversal is possible, and enumeration of
                columns is indicative of available destination nodes.

        """
        out = np.zeros((len(self.node_map), len(self.node_map)))

        for i, (node_parent, node_children) in enumerate(self.link_map.items()):
            for j, (node_child, _) in enumerate(self.link_map.items()):
                if i == j:
                    continue
                elif node_child in node_children:
                    out[i, j] = 1

        return out

    @property
    def node_sources(self) -> list:
        """
        Returns:
            A list of nodes from which other (non-source) nodes are downstream.

        """
        indexed_nodes = {i: k for i, (k, _) in enumerate(self.node_map.items())}
        sources = [
            indexed_nodes[i]
            for i, arr in enumerate(self.mat_adjacency.T)
            if (arr == 0).all()  # No inbound directed paths
        ]
        return sources

    @property
    def node_sinks(self) -> list:
        """
        Returns:
            A list of nodes from which other (non-sink) nodes are upstream.

        """
        indexed_nodes = {i: k for i, (k, _) in enumerate(self.node_map.items())}
        sinks = [
            indexed_nodes[i]
            for i, arr in enumerate(self.mat_adjacency)
            if (arr == 0).all()  # No outbound directed paths
        ]
        return sinks

    @property
    def flow(self) -> list:
        """
        Returns:
            The order with which nodes must be computed to ensure downstream
                flow of information.

        """
        # Initialise...
        flow_order = []  # means of recording indices of nodes in flow order
        adjacency = self.mat_adjacency

        while ~np.isnan(adjacency).all():
            queue = []
            for i, node_adj in enumerate(adjacency):
                if (node_adj[~np.isnan(node_adj)] == 0).all():  # If no children
                    if i not in flow_order:  # If not already designated in flow
                        queue.append(i)
            else:
                for i in queue:
                    flow_order.insert(0, i)
                    adjacency[i] = np.nan  # Invalidate the parent
                    adjacency[:, i] = np.nan  # Invalidate connections of parent

            # The queue was empty, flow is not possible
            if not queue:
                errormsg = "Graph is disconnected or cyclic, flow not possible"
                raise StopIteration(errormsg)

        # Convert reference flow indices into the respective nodes of the graph
        indexed_nodes = {i: k for i, (k, _) in enumerate(self.node_map.items())}
        return [indexed_nodes[i] for i in flow_order]


class Graphs(object):
    """A collection of tools for creating and analyzing (networked) graphs."""
    Graph = Graph
    Traverse = Traverse



"""Module implementing means of accessing and parsing weather reports."""
import time
import warnings

from metar.Metar import Metar
import numpy as np
import requests

from carpy.utility import call_count

__all__ = ["METARtools"]
__author__ = "Yaseen Reza"


@call_count
def _metar_from_api(*station_ids: str, hours: int = None) -> str:
    """
    Return a string of historical METAR observations at desired stations. This
    is achieved by querying 'https://aviationweather.gov'. Calls to this
    function are not rate limited, so please use responsibly (ideally not
    at all if you are an end-user).

    Args:
        *station_ids: String arguments of ICAO station IDs to query
        hours: Integer number of hours from which to pull historical METAR
            records. Optional, defaults to 3 hour.

    Returns:
        METAR string with reports separated by the newline character.

    Examples: ::

        # Get last 2 hrs of data for London Heathrow and Southampton Airport
        >>> mymetars = _metar_from_api("EGLL", "EGHI", hours=2)

        # Read the first line of the metar string
        >>> print(mymetars.splitlines()[0])
        'EGHI 270850Z 12005G15KT 090V190 9999 BKN025 11/06 Q1015'

    """
    # Recast as necessary
    hours = int(3 if hours is None else hours)

    # Build request
    url_base = "https://beta.aviationweather.gov/"
    api_path = "cgi-bin/data/metar.php?ids="
    api_querytext = ",".join(station_ids) + f"&hours={hours}"

    # Perform query
    try:
        link = f"{url_base}{api_path}{api_querytext}"
        response = requests.get(url=link)
    except Exception as _:
        errormsg = "Couldn't access weather service. Is your device connected?"
        raise ConnectionError(errormsg)

    if response.status_code != 200:
        errormsg = f"Couldn't access METAR data ({response.status_code=})"
        raise ConnectionError(errormsg)

    # Return results of a successful query
    return response.text


class METARtools(object):
    """A collection of tools for working with METAR observations."""

    @staticmethod
    def fetch(*station_ids: str, hours: int = None) -> list:
        """
        Return a string of historical METAR observations at desired stations.
        Calls to the API embedded in this method are rate limited to prevent
        and discourage spam to the API host.

        Args:
            *station_ids: String arguments of ICAO station IDs to query
            hours: Integer number of hours from which to pull historical METAR
                records. Optional, defaults to 3 hour.

        Returns:
            METAR string with reports separated by the newline character.

        Examples: ::

            # Get last 2 hrs of data for London Heathrow and Southampton Airport
            >>> mymetars = METARtools.fetch("EGLL", "EGHI", hours=2)

            # Read the first line of the metar string
            >>> print(mymetars[0])
            'EGHI 270850Z 12005G15KT 090V190 9999 BKN025 11/06 Q1015'

        """
        # Rate-limit calls to the API according to a logarithmic schedule
        api_response = _metar_from_api(*station_ids, hours=hours)
        api_calls = _metar_from_api.call_count
        warnmsg = (
            f"This method depends on the functionality of an API service that "
            f"is currently in beta and may break without warning. Notify the "
            f"developers if this feature is not working as expected!"
        )
        if api_calls == 1:
            warnings.warn(message=warnmsg, category=RuntimeWarning)
        time.sleep(np.log(api_calls + 1))

        # Let's be real, most users want a list of METARs, not a single string
        metarlist = api_response.splitlines()
        return metarlist

    @staticmethod
    def decode(*METARs: str) -> list:
        """
        Return decoded METAR observations.

        Args:
            *METARs: String arguments of coded METAR observations.

        Returns:
            A list of Metar objects that corresponding to the argument METARs.

        """
        decoded_obs = [Metar(metar_str) for metar_str in METARs]
        return decoded_obs



from carpy.utility import Graphs


class NetWork(object):
    """
    The NetWork class is used to define a network of components constituent to a
    vehicle's propulsion system, in service of a goal to evaluate the vehicle's
    gross thrust (which is related to the rate of net work).
    """

    def __init__(self):
        self._graph = Graphs.Graph()
        return

    @property
    def graph(self) -> Graphs.Graph:
        """
        Returns:
            Graph object, for mapping nodes (vertices) and links (edges).

        """
        return self._graph

    def compile(self):
        visited = self.graph.node_sources
        flow_order = self.graph.flow  # Compute order of nodes

        # Identify nodes that are missing power distribution estimates
        nodes2compute = flow_order[len(visited):]

        for node in nodes2compute:
            neighbours = self.graph.node_map[node]
            print(node, neighbours)

        return


if __name__ == "__main__":
    from carpy.propulsion.components._power_raise import PhotovoltaicPanel, \
        Battery, PSU

    mynetwork = NetWork()
    panelPV1 = mynetwork.graph.new_node(PhotovoltaicPanel())
    panelPV2 = mynetwork.graph.new_node(PhotovoltaicPanel())
    panelPV3 = mynetwork.graph.new_node(PhotovoltaicPanel())
    PSU = mynetwork.graph.new_node(PSU(eta=0.89))
    battery1 = mynetwork.graph.new_node(Battery(specific_energy=80))
    battery2 = mynetwork.graph.new_node(Battery(specific_energy=80))

    mynetwork.graph.new_link(panelPV1, PSU, directed=True)
    mynetwork.graph.new_link(panelPV2, PSU, directed=True)
    mynetwork.graph.new_link(panelPV3, PSU, directed=True)

    mynetwork.graph.new_link(PSU, battery1, directed=True)
    mynetwork.graph.new_link(PSU, battery2, directed=True)

    mynetwork.compile()



"""Tests for library maths methods."""
import unittest

import numpy as np

from carpy.utility import Quantity, interp_lin, interp_exp


class LinearInterpolation(unittest.TestCase):
    """Check the linear interpolator works okay."""

    def test_float(self):
        """Check the linear interpolator can handle a single scalar."""
        interp = interp_lin(0.2, [0, 1], [0, 100])
        self.assertEqual(20, interp, "Failed to interpolate scalar input")
        return

    def test_array(self):
        """Check the linear interpolator can handle a vector array."""
        interp = interp_lin([0.2, 0.3], [0, 1], [0, 100])
        self.assertTrue(all(np.array([20, 30]) == interp))
        return

    def test_quantity(self):
        """Check the linear interpolator can handle quantities."""
        interp = interp_lin([0.2, 0.3], [0, 1], Quantity([0, 100], "kg"))
        # Check value
        self.assertTrue(all(Quantity([20, 30], "kg") == interp))
        # Check type
        self.assertIsInstance(interp, Quantity)
        return


class ExponentialInterpolation(unittest.TestCase):

    def test_float(self):
        """Check the exponential interpolator can handle a single scalar."""
        interp = interp_exp(0.5, [0, 1], np.exp([0, 1]))
        self.assertTrue(np.isclose(np.exp(0.5), interp, atol=1e-5))
        return

    def test_array(self):
        """Check that the exponential interpolator can handle a vector array."""
        interp = interp_exp([0.5, 0.75], [0, 1], np.exp([0, 1]))
        self.assertTrue(all(np.isclose(np.exp([0.5, 0.75]), interp, atol=1e-5)))
        return

    def test_quantity(self):
        """Check that the exponential interpolator can handle quantities."""
        interp = interp_exp([0.5, 0.75], [0, 1], Quantity(np.exp([0, 1]), "kg"))
        # Check value
        self.assertTrue(
            all(np.isclose(np.exp([0.5, 0.75]), np.array(interp), atol=1e-5)))
        # Check type
        self.assertIsInstance(interp, Quantity)
        return


if __name__ == '__main__':
    unittest.main()



"""Methods for generating and optimising wing planforms."""
import numpy as np

from carpy.aerodynamics.aerofoil import Aerofoil
from carpy.structures import DiscreteIndex
from carpy.utility import Hint, Quantity, collapse_array, isNone

__all__ = ["WingSection", "WingSections"]
__author__ = "Yaseen Reza"


class WingSection(object):
    """
    Class for modelling wing cross-sections, as aligned with wing buttock lines.

    A wing section is a 2D cross-sectional slice of the 3D wing structure, and
    inclined at the local angle of dihedral. Sections are defined through a
    non-dimensional aerofoil geometry, the chord length, and the applied twist,
    sweep, and dihedral angles.
    """

    def __init__(self, aerofoil: Aerofoil = None, chord: Hint.num = None,
                 twist: Hint.num = None):
        self._aerofoil = aerofoil
        self._twist = 0.0 if twist is None else float(twist)
        self.chord = 1.0 if chord is None else float(chord)
        self._sweep = None
        self._dihedral = None
        return

    def __add__(self, other):
        # Typechecking
        if not isinstance(other, type(self)):
            raise TypeError(f"Cannot add {type(self)=} to {type(other)=}")

        new_object = type(self)(
            aerofoil=self.aerofoil + other.aerofoil,
            chord=self.chord + other.chord,
            twist=self.twist + other.twist
        )
        return new_object

    def __mul__(self, other):
        # Typechecking
        if not isinstance(other, Hint.num.__args__):
            raise TypeError(f"Cannot multiply {type(self)=} by {type(other)=}")
        # New non-dimensional profile, chord length, and angle of twist
        new_aerofoil = other * self.aerofoil
        new_chord = other * self.chord
        new_theta = other * self.twist

        new_object = type(self)(
            aerofoil=new_aerofoil,
            chord=new_chord,
            twist=new_theta
        )
        return new_object

    def deepcopy(self):
        """Returns a deep copy of self (no values are shared in memory)."""
        return type(self)(aerofoil=self._aerofoil, twist=self._twist)

    @property
    def aerofoil(self):
        """Non-dimensional aerofoil object, attached to the station."""
        return self._aerofoil

    @property
    def chord(self) -> Quantity:
        """
        The chord length of the station's aerofoil.

        Returns:
            Reference chord length of the aerofoil at this wing station.

        """
        return self._chord

    @chord.setter
    def chord(self, value):
        self._chord = Quantity(value, "m")
        return

    @property
    def twist(self) -> float:
        """
        The station's geometric angle of twist, relative to wing root. For a
        horizontal wing, a positive value indicates wash-in (station has an
        angle of incidence greater than the wing root), and a negative value
        indicates wash-out.

        Returns:
            Geometric angle of twist of the wing station.

        """
        return self._twist

    @twist.setter
    def twist(self, value):
        self._twist = float(value)
        return

    @property
    def sweep(self) -> float:
        """
        The sweep angle of the leading edge. A positive value indicates sweep in
        the traditional sense (outboard leading edge is behind inboard leading
        edge), and a negative value indicates forward- or reverse-sweep.

        Returns:
            Geometric angle of sweep of the wing station's leading edge.

        """
        return self._sweep

    @sweep.setter
    def sweep(self, value):
        self._sweep = float(value)
        return

    @property
    def dihedral(self) -> float:
        """
        The angle between the station plane's normal vector and the vehicle
        plane formed of longitudinal (X) and lateral (Y) axes.

        Returns:
            Geometric angle of dihedral of the wing station.

        """
        return self._dihedral

    @dihedral.setter
    def dihedral(self, value):
        self._dihedral = float(value)
        return


class WingSections(DiscreteIndex):

    def __init__(self):
        # Super class call
        super().__init__()

        return

    def __getitem__(self, key):
        nd_sections = super().__getitem__(key)

        if isinstance(key, slice):
            for slice_bound in [key.start, key.stop]:
                if slice_bound not in self and not isNone(slice_bound):
                    errormsg = (
                        f"Slice is bounded by key that has no associated value."
                        f" Try creating a new element of {self} with the index "
                        f"{slice_bound}"
                    )
                    raise RuntimeError(errormsg)

        # Cast to list temporarily, if necessary
        if not isinstance(nd_sections, Hint.iter.__args__):
            nd_sections = [nd_sections]

        # Missing sweep and dihedral angles should be inherited from inboard
        for i, aerofoil in enumerate(nd_sections):
            # If aerofoil is derived, it has parents (one of which is inboard)
            if hasattr(aerofoil, "_parents"):
                parent = self[min(getattr(aerofoil, "_parents"))]
                # Assign private vars, skips float typecasting for 'None' values
                aerofoil._sweep = parent.sweep
                aerofoil._dihedral = parent.dihedral
                delattr(aerofoil, "_parents")

        # If sliced, return an array
        if isinstance(key, slice):
            return nd_sections
        return collapse_array(nd_sections)


if __name__ == "__main__":
    from carpy.aerodynamics.aerofoil import NewAerofoil

    n0012 = NewAerofoil.from_method.NACA("0012")
    n8412 = NewAerofoil.from_method.NACA("8412")

    # Define buttock-line geometry
    mysections = WingSections()
    mysections[0] = WingSection(n8412)
    mysections[60] = mysections[0].deepcopy()
    mysections[100] = WingSection(n0012)

    # Add leading edge sweep and dihedral
    mysections[:60].sweep = np.radians(0)
    mysections[60:].sweep = np.radians(2)
    mysections[0:].dihedral = np.radians(3)

    # Introduce wing taper
    mysections[0:].chord = 1.0
    mysections[100].chord = 0.4

# class WingPlane(object):
#     """
#     Class for modelling wing planes.
#     """
#
#     def __init__(self, span: Hint.num, mirror: bool = None):
#         """
#         Args:
#             span: The full span of the wing, as determined from planform view.
#             mirror: Whether or not to mirror the wing's stations about the
#                 centreline.
#         """
#         self._b = span
#         self._mirror = True if mirror is None else mirror
#         self._spar = None
#
#         return
#
#     @property
#     def parametric_spar(self) -> Hint.func:
#         """
#         A parameterised definition of the wingspar's (composite) geometry and
#         material selection(s). Accepts arguments for the maximal 'height' and
#         'width' of a spar section.
#
#         Returns:
#             Spar section object, with geometric properties pre-computed.
#
#         """
#         if self._spar is None:
#             raise NotImplementedError("Spar definitions has not yet been given")
#         return self._spar
#
#     @parametric_spar.setter
#     def parametric_spar(self, value):
#         if not callable(value):
#             errormsg = (
#                 f"section_spar.setter is expecting to be given 'function(y)', "
#                 f"actually got section_spar = {value} (invalid {type(value)=})"
#             )
#             raise TypeError(errormsg)
#         self._spar = value
#         return
#
#     @parametric_spar.deleter
#     def parametric_spar(self):
#         self._spar = None
#         return

if __name__ == "__main__":
    from sectionproperties.pre.library import steel_sections, primitive_sections
    from sectionproperties.pre.pre import Material
    from sectionproperties.analysis.section import Section

    steel = Material(
        name='Steel', elastic_modulus=200e9, poissons_ratio=0.3,
        density=7.85e3, yield_strength=500e6, color='grey'
    )
    timber = Material(
        name='Timber', elastic_modulus=8e9, poissons_ratio=0.35,
        density=6.5e2, yield_strength=20e6, color='burlywood'
    )


    def sections(height: float, width: float):
        """

        Args:
            height: Maximum allowable height of the spar.
            width: Maximum allowable width of the spar.

        Returns:

        """
        # Compute parameterised dimensions
        core_y = height - (2 * width)
        tube_od = width
        tube_wt = 1e-3

        # STEP 1: Create component section geometries (and attach materials)
        rod_kwargs = {"d": tube_od, "t": tube_wt, "n": 100, "material": steel}
        core_kwargs = {"b": rod_kwargs["d"], "d": core_y, "material": timber}
        rod = steel_sections.circular_hollow_section(**rod_kwargs)
        core = primitive_sections.rectangular_section(**core_kwargs)

        # STEP 2: Create a compound geometry, and convert into a section object
        section_geometry = (
                rod.shift_section(0, (core_y + rod_kwargs["d"]) / 2)
                + core.shift_section(-rod_kwargs["d"] / 2, -core_y / 2)
                + rod.shift_section(0, -(core_y + rod_kwargs["d"]) / 2)
        )
        section_geometry.create_mesh(mesh_sizes=[1e-3])
        section = Section(section_geometry)

        # STEP 3: Calculate the geometric properties of the section
        section.calculate_geometric_properties()

        return section


    section = sections(height=0.1, width=12e-3)
    section.plot_mesh()



"""Module implementing models of the Earth's atmosphere."""
import functools
import itertools
import os
import re
import warnings

import numpy as np
import pandas as pd

from carpy.gaskinetics import GasModels, IsentropicFlow
from carpy.utility import (
    GetPath, Hint, Pretty, Quantity,
    cast2numpy, cast2quantity, constants as co, isNone, interp_lin, interp_exp
)

__all__ = ["ISA1975", "US1976", "MILHDBK310", "ObsAtmospherePerfect"]
__author__ = "Yaseen Reza"


# ============================================================================ #
# Load data required by the module
# ---------------------------------------------------------------------------- #
def load_iso2533(filename: str, /):
    """Load and return dataframes for the ISO 2533:1975 standard."""
    filepath = os.path.join(GetPath.localpackage(), "data", filename)
    dataframes = pd.read_excel(io=filepath, sheet_name=None, na_filter=False)
    return dataframes


df_ISA1975 = load_iso2533("ISO-2533-1975.xlsx")


def load_milhdbk310(filename: str, /):
    """Load and return dataframes for the MIL-HDBK-310 specification."""
    filepath = os.path.join(GetPath.localpackage(), "data", filename)

    def read_as_number(x):
        """Read and parse numbers as they are given in the MIL-HDBK-310 spec."""
        x = str(x)
        components = re.findall(r"\d+\.?(?:\d+)?|[+-]\d", x)
        if len(components) == 1:
            return float(components[0])
        else:
            return eval(f"{components[0]}e{components[1]}")

    columns = pd.read_excel(io=filepath, nrows=0).columns
    dataframes = pd.read_excel(io=filepath, sheet_name=None, na_filter=False,
                               converters={k: read_as_number for k in columns})
    return dataframes


df_MILHDBK310 = load_milhdbk310("MIL-HDBK-310_profiles.xlsx")


# ============================================================================ #
# Support functions
# ---------------------------------------------------------------------------- #

def atmosphere_builder(altitude: Hint.nums, T: Hint.nums = None,
                       p: Hint.nums = None, rho: Hint.nums = None,
                       Mbar: float = None):
    """
    Return functions for temperature, pressure, and density with altitude.

    Given data for altitude and any valid combinations of temperature, pressure,
    and density at the associated altitude levels, return functions that accept
    altitude as an argument and return the corresponding values for T, p, and
    rho.

    Args:
        altitude: Altitude data. Assumed to be increasing.
        T: Temperature data, in the same shape as the altitude data. Optional.
        p: Pressure data, in the same shape as the altitude data. Optional.
        rho: Density data, in the same shape as the altitude data. Optional.
        Mbar: Average molar mass of atmospheric particles. Optional.

    Returns:
        f_temperature(altitude), f_pressure(altitude), f_density(altitude).

    """
    # Recast as necessary
    altitude = Quantity(altitude, "m")
    T = None if T is None else Quantity(T, "K")
    p = None if p is None else Quantity(p, "Pa")
    rho = None if rho is None else Quantity(rho, "kg m^{-3}")
    Mbar = co.STANDARD.SL.M if Mbar is None else Quantity(Mbar, "kg mol^{-1}")
    # pre-package kwargs for interpolation and extrapolation of atm. quantities
    commonkwargs = dict([("xp", altitude), ("bounded", True)])

    # Verify that altitude data is increasing
    if all(np.diff(altitude) <= 0):
        errormsg = f"Elements of 'altitude' data must be increasing"
        raise ValueError(errormsg)

    # If temperature is provided and no other arguments are present
    if T is not None and all(isNone(p, rho)):

        # Because pressure is not given, we must assume that an altitude of zero
        # corresponds with standard sea-level pressure. We have to make sure
        # zero appears in the altitude array (with corresponding temperature).
        idx_of_SL = len(altitude[altitude < 0])
        if 0 not in altitude:
            # Make sure 0 appears in the altitude listing and recast temperature
            altitude_with_zero = np.insert(altitude, idx_of_SL, 0)
            T = interp_lin(altitude_with_zero, altitude, T)
            altitude = altitude_with_zero

        # Initialise base-layer pressure
        p0 = co.STANDARD.SL.p
        T_lapserate = np.diff(T) / np.diff(altitude)
        for i in range(idx_of_SL):
            ui = idx_of_SL - i
            li = idx_of_SL - (i + 1)
            daltitude = altitude[ui] - altitude[li]

            # Molecular-scale pressure equation depends on lapse rate
            if T_lapserate[i] != 0:
                exponent = np.asarray(
                    co.STANDARD.SL.g / co.STANDARD.SL.Rs / T_lapserate[li])
                p0 /= (T[li] / (T[ui])) ** exponent
            else:
                exponent = np.asarray(
                    -co.STANDARD.SL.g * daltitude / co.STANDARD.SL.Rs / T[li])
                p0 /= np.exp(exponent)

        # Define functions
        f_T = functools.partial(interp_lin, fp=T, **commonkwargs)

        def f_p(y):
            """Atmospheric pressure as a function of altitude."""
            # Initialise values to lowest altitude
            Ts = np.ones(y.shape) * T[0]
            ps = np.ones(y.shape) * p0

            for j in range(len(altitude) - 1):
                # Clip alt. by the layer's upper bound, and diff. to lower bound
                # If you're a future developer looking to find out how to
                # extrapolate atmospheric pressure modelling, this is what's
                # limiting the atmospheric model.
                dy = np.clip(
                    np.clip(y, None, altitude[j + 1]) - altitude[j],
                    0, None
                )
                # Molecular-scale pressure equation depends on lapse rate
                if T_lapserate[j] != 0:
                    ps *= (Ts / (Ts + T_lapserate[j] * dy)) ** np.asarray(
                        co.STANDARD.SL.g / co.STANDARD.SL.Rs / T_lapserate[j])
                else:
                    ps *= np.exp(np.asarray(
                        -co.STANDARD.SL.g * dy / co.STANDARD.SL.Rs / Ts))

                # Update temperature after the fact, as we needed temperature
                # from the base of the layer to make the computation
                Ts += T_lapserate[j] * dy

            return ps

        def f_rho(y):
            """Compute density, given altitude"""
            return f_p(y) * Mbar / co.PHYSICAL.R / f_T(y)

    elif not all(isNone(T, p)) and rho is None:
        f_T = functools.partial(interp_lin, fp=T, **commonkwargs)
        f_p = functools.partial(interp_exp, fp=p, **commonkwargs)

        def f_rho(y):
            """Compute density, given altitude"""
            return f_p(y) * Mbar / co.PHYSICAL.R / f_T(y)

    elif not all(isNone(T, rho)) and p is None:
        f_T = functools.partial(interp_lin, fp=T, **commonkwargs)
        f_rho = functools.partial(interp_exp, fp=rho, **commonkwargs)

        def f_p(y):
            """Compute pressure, given altitude"""
            return f_rho(y) / (Mbar / co.PHYSICAL.R / f_T(y))

    else:
        errormsg = "Got insufficient data to generate p, T, and rho functions"
        raise ValueError(errormsg)

    return f_T, f_p, f_rho


def cast_altitudes(f_geometric: bool):
    """
    Create and return a decorator that permits use of geometric or gravitational
    potential altitudes with atmospheric property models.

    Args:
        f_geometric: Boolean flag to indicate whether the function being
            decorated is defined in terms of geometric or potential altitude.
    """

    def decorator(method):
        """Create, apply, and return the method's wrapper."""

        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            """Augment atmospheric modelling functions with default behaviour"""
            # Find all arguments
            all_args = args + tuple(v for (k, v) in kwargs.items())

            # Make sure positional argument is given
            instance, altitude = all_args[0:2]
            if altitude is None:
                errormsg = (
                    f"{type(instance).__name__}.{method.__name__}() missing 1 "
                    f"required positional argument: 'altitude'"
                )
                raise TypeError(errormsg)

            # Recast as necessary
            altitude = cast2numpy(altitude)
            if len(args) < 3:
                geometric = kwargs.get("geometric", None)
                geometric = False if geometric is None else geometric
            elif len(args) == 3:
                geometric = args[2]
            else:
                errormsg = (
                    f"{type(instance).__name__}.{method.__name__}() takes 2 "
                    f"positional arguments but {len(args) - 1} were given"
                )
                raise TypeError(errormsg)

            # Function uses geometric inputs, and user specified geometric input
            if f_geometric and geometric:
                return method(altitude)
            # Function uses geometric inputs, and user specified potential input
            elif f_geometric and not geometric:
                return method(instance.geop2geom(altitude))
            # Function uses potential inputs, and user specified geometric input
            elif not f_geometric and geometric:
                return method(instance.geom2geop(altitude))
            # Function uses potential inputs, and user specified potential input
            else:
                return method(altitude)

        return wrapper

    return decorator


# ============================================================================ #
# Atmosphere objects
# ---------------------------------------------------------------------------- #
# Base objects

class Atmosphere(object):
    """Class for models of the Earth's atmosphere based on perfect gases."""

    _f_T = NotImplemented
    _f_p = NotImplemented
    _f_rho = NotImplemented
    _f_c_sound = NotImplemented
    _f_mu_visc = NotImplemented
    _f_k_thermal = NotImplemented
    _f_gamma = NotImplemented
    _r_Earth = co.STANDARD.ISO_2533.r  # Default nominal radius of Earth

    def __init__(self, T_offset: float = None):
        self._T_offset = float(0 if T_offset is None else T_offset)
        return

    def __repr__(self):
        reprstring = f"{type(self).__name__}(Toffset={self.T_offset})"
        return reprstring

    def __str__(self):
        rtnstring = (
            f"{type(self).__name__}({Pretty.temperature(self.T_offset)})")
        return rtnstring

    @functools.cached_property
    def T_offset(self) -> Quantity:
        """
        Temperature offset compared to the original atmosphere model. A T_offset
        of -10 degrees Celsius (-10 Kelvin) indicates that the atmosphere is 10
        degrees cooler than usual.

        Returns:
            Atmospheric temperature offset.

        """
        return Quantity(self._T_offset, "K")

    @classmethod
    def geom2geop(cls, altitude: Hint.nums) -> Quantity:
        """
        Compute gravitational potential altitude given the geometric altitude.

        Args:
            altitude: Geometric altitude(s) to be converted.

        Returns:
            Geopotential altitude(s).

        """
        # Recast as necessary
        geom = Quantity(altitude, "m")

        # Compute geopotential altitude
        geop = cls._r_Earth * geom / (cls._r_Earth + geom)

        return geop

    @classmethod
    def geop2geom(cls, altitude: Hint.nums) -> Quantity:
        """
        Compute geometric altitude given the gravitational potential altitude.

        Args:
            altitude: Gravitational potential altitude(s) to be converted.

        Returns:
            Geometric altitude(s).

        """
        # Recast as necessary
        geop = Quantity(altitude, "m")

        # Compute geometric altitude
        geom = cls._r_Earth * geop / (cls._r_Earth - geop)

        return geom

    @classmethod
    def g_acc(cls, altitude: Hint.nums, phi: Hint.nums = None) -> Quantity:
        """
        Return the acceleration due to gravity.

        Args:
            altitude: The altitude above mean sea level.
            phi: Latitude of the query. Optional, defaults to 45 32' 33" degrees
                North.

        Returns:
            Acceleration due to gravity.

        """
        # Recast as necessary
        altitude = cast2numpy(altitude)
        # 45 32'33" North
        phi = cast2numpy(np.radians(45.5425) if phi is None else phi)

        # Lambert's equation for acceleration of free-fall with latitude
        cosphi2 = np.cos(2 * phi)
        g_phi = Quantity(
            9.806_16 * (1 - 0.002_637_3 * cosphi2 + 0.000_005_9 * cosphi2 ** 2),
            "m s^{-2}"
        )

        # Lapse of acceleration with altitude
        g = g_phi * cls._r_Earth ** 2 / (cls._r_Earth + altitude) ** 2

        return g

    def T(self, altitude: Hint.nums, geometric: bool = False) -> Quantity:
        """
        Return the outside air temperature with altitude.

        Args:
            altitude: Altitude at which to query the atmosphere model. By
                default, the altitude is considered geopotentially scaled
                (unless otherwise specified using the 'geometric' argument).
            geometric: Flag specifying if given altitudes are geometric or not.
                Optional, defaults to False.

        Returns:
            Outside air temperature.

        """
        offset = self.T_offset
        return Quantity(self._f_T(self, altitude, geometric) + offset, "K")

    def p(self, altitude: Hint.nums, geometric: bool = False) -> Quantity:
        """
        Return the outside static air pressure with altitude.

        Args:
            altitude: Altitude at which to query the atmosphere model. By
                default, the altitude is considered geopotentially scaled
                (unless otherwise specified using the 'geometric' argument).
            geometric: Flag specifying if given altitudes are geometric or not.
                Optional, defaults to False.

        Returns:
            Outside static air pressure.

        """
        return Quantity(self._f_p(self, altitude, geometric), "Pa")

    def rho(self, altitude: Hint.nums, geometric: bool = False) -> Quantity:
        """
        Return the outside air density with altitude.

        Args:
            altitude: Altitude at which to query the atmosphere model. By
                default, the altitude is considered geopotentially scaled
                (unless otherwise specified using the 'geometric' argument).
            geometric: Flag specifying if given altitudes are geometric or not.
                Optional, defaults to False.

        Returns:
            Outside air density.

        """
        _ = geometric  # A "do nothing" statement, geometric is for the wrapper
        return Quantity(self._f_rho(self, altitude, geometric), "kg m^{-3}")

    def gamma(self, altitude: Hint.nums, geometric: bool = False) -> Quantity:
        """
        Return the ratio of specific heats of air with altitude.

        Args:
            altitude: Altitude at which to query the atmosphere model. By
                default, the altitude is considered geopotentially scaled
                (unless otherwise specified using the 'geometric' argument).
            geometric: Flag specifying if given altitudes are geometric or not.
                Optional, defaults to False.

        Returns:
            Adiabatic index of air.

        """
        return self._f_gamma(altitude, geometric)

    def c_sound(self, altitude: Hint.nums, geometric: bool = False) -> Quantity:
        """
        Return the speed of sound in air with altitude.

        Args:
            altitude: Altitude at which to query the atmosphere model. By
                default, the altitude is considered geopotentially scaled
                (unless otherwise specified using the 'geometric' argument).
            geometric: Flag specifying if given altitudes are geometric or not.
                Optional, defaults to False.

        Returns:
            Speed of sound in air.

        """
        return Quantity(self._f_c_sound(altitude, geometric), "m s^{-1}")

    def mu_visc(self, altitude: Hint.nums, geometric: bool = False) -> Quantity:
        """
        Return the dynamic viscosity of air with altitude.

        Args:
            altitude: Altitude at which to query the atmosphere model. By
                default, the altitude is considered geopotentially scaled
                (unless otherwise specified using the 'geometric' argument).
            geometric: Flag specifying if given altitudes are geometric or not.
                Optional, defaults to False.

        Returns:
            Dynamic viscosity of air.

        """
        return Quantity(self._f_mu_visc(altitude, geometric), "Pa s")

    def _f_nu(self, altitude: Hint.nums, geometric: bool = False) -> Quantity:
        """Return the kinematic viscosity of air with altitude."""
        # Take advantage of the relationship with dynamic viscosity
        kwargs = {"altitude": altitude, "geometric": geometric}
        return self._f_mu_visc(**kwargs) / self._f_rho(**kwargs)

    def nu_visc(self, altitude: Hint.nums, geometric: bool = False) -> Quantity:
        """
        Return the kinematic viscosity of air with altitude.

        Args:
            altitude: Altitude at which to query the atmosphere model. By
                default, the altitude is considered geopotentially scaled
                (unless otherwise specified using the 'geometric' argument).
            geometric: Flag specifying if given altitudes are geometric or not.
                Optional, defaults to False.

        Returns:
            Kinematic viscosity of air.

        """
        # Take advantage of the relationship with dynamic viscosity
        kwargs = {"altitude": altitude, "geometric": geometric}
        return self._f_mu_visc(**kwargs) / self._f_rho(**kwargs)

    def k_thermal(self, altitude: Hint.nums,
                  geometric: bool = False) -> Quantity:
        """
        Return the thermal conductivity of air with altitude.

        Args:
            altitude: Altitude at which to query the atmosphere model. By
                default, the altitude is considered geopotentially scaled
                (unless otherwise specified using the 'geometric' argument).
            geometric: Flag specifying if given altitudes are geometric or not.
                Optional, defaults to False.

        Returns:
            Thermal conductivity of air.

        """
        return self._f_k_thermal(altitude, geometric)

    def airspeeds(self, altitude: Hint.nums = None, geometric=False, *,
                  CAS: Hint.nums = None, EAS: Hint.nums = None,
                  TAS: Hint.nums = None, Mach: Hint.nums = None) -> tuple:
        """
        Return calibrated, equivalent, and true airspeed at the query altitude.

        Args:
            altitude: Altitude at which to query the atmosphere model. By
                default, the altitude is considered geopotentially scaled
                (unless otherwise specified using the 'geometric' argument).
            geometric: Flag specifying if given altitudes are geometric or not.
                Optional, defaults to False.
            CAS: Calibrated (indicated) airspeed.
            EAS: Equivalent airspeed at sea level.
            TAS: True airspeed of the fluid.
            Mach: The Mach number of flight.

        Returns:
            tuple: (CAS, EAS, TAS, M) in metres per second (where applicable).

        """
        # Recast inputs as necessary
        altitude = cast2numpy(altitude)
        kwargs = {"altitude": altitude, "geometric": geometric}
        CAS = CAS if CAS is None else Quantity(CAS, "m s^{-1}")
        EAS = EAS if EAS is None else Quantity(EAS, "m s^{-1}")
        TAS = TAS if TAS is None else Quantity(TAS, "m s^{-1}")
        Mach = Mach if Mach is None else cast2numpy(Mach)

        # Verify validity of inputs, and that only one input is given
        if tuple(isNone(CAS, EAS, TAS, Mach)).count(True) == 3:
            pass
        else:
            errormsg = (
                f"Expected one of CAS, EAS, TAS, and Mach arguments should be "
                f"used (got {CAS=}, {EAS=}, {TAS=}, {Mach=})"
            )
            raise ValueError(errormsg)

        # Solve for the missing speeds!
        rho = self.rho(**kwargs)
        rho_sl = self.rho(altitude=0)
        p = self.p(**kwargs)
        p_sl = self.p(altitude=0)
        a = self.c_sound(**kwargs)
        a_sl = self.c_sound(altitude=0)
        gamma = self.gamma(**kwargs)
        while True:
            # EAS can be computed from TAS
            if EAS is None and TAS is not None:
                EAS = TAS * np.sqrt(rho / rho_sl)
            # TAS can be computed from EAS
            elif TAS is None and EAS is not None:
                TAS = EAS / np.sqrt(rho / rho_sl)
            # CAS can be computed from TAS
            elif CAS is None and EAS is not None and TAS is not None:
                # Use isentropic flow relations to find the Mach number
                Mach = TAS / a
                pt_p = 1 / IsentropicFlow.p_p0(M=Mach, gamma=gamma)
                qc = p * (pt_p - 1)
                Tt_Tsl = (qc / p_sl + 1) ** ((gamma - 1) / gamma)
                Tt_T = (qc / p + 1) ** ((gamma - 1) / gamma)
                # If Mach is zero, there is a divide by zero error. Suppress it.
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    CAS = EAS * ((p_sl / p) * (Tt_Tsl - 1) / (Tt_T - 1)) ** 0.5
                CAS = np.where(np.isnan(CAS), 0, CAS)  # replace the nans with 0
            # TAS can be computed from CAS
            elif TAS is None and CAS is not None:
                # Use isentropic flow relations to find the Mach number
                M_sl = CAS / a_sl
                pt_psl = 1 / IsentropicFlow.p_p0(M=M_sl, gamma=gamma)
                qc = p_sl * (pt_psl - 1)
                Mach = IsentropicFlow.M(p_p0=(p / (p + qc)), gamma=gamma)
                TAS = Mach * a
            # TAS can be computed from Mach
            elif TAS is None and Mach is not None:
                TAS = Mach * a
            # It is only possible to reach this stage if we have all the params
            else:
                break  # Leave the while loop

        # Find the shape of the output array by finding the biggest shape
        oshape = max([
            x.shape for x in [altitude, CAS, EAS, TAS, Mach] if x is not None])
        # These speed arrays are flattened, 1D and need to be cast back to shape
        speeds = np.array(np.broadcast_arrays(CAS, EAS, TAS, Mach))

        # Cast back into the right shape
        speeds = tuple([
            (Quantity(x, "m s^{-1}") if i != 3 else x).reshape(oshape)
            for i, x in enumerate(speeds)
        ])
        return speeds


# ---------------------------------------------------------------------------- #
# International Standard Atmosphere (1975)

class ISA1975(Atmosphere):
    """
    Simple model of the International Standard Atmosphere, as given in the
    ISO 2533:1975 standard.
    """

    def __init__(self, T_offset: float = None):
        # Super call
        super().__init__(T_offset=T_offset)

        # Get fundamental atmopsheric properties (in terms of geopotential alt.)
        table4 = cast2quantity(df_ISA1975["Table4"])
        f_temperature, f_pressure, f_density = atmosphere_builder(
            altitude=table4["Geopotential altitude, H"],
            T=table4["Temperature, T"],
            Mbar=co.STANDARD.ISO_2533.M
        )
        self._f_T = cast_altitudes(f_geometric=False)(f_temperature)
        self._f_p = cast_altitudes(f_geometric=False)(f_pressure)
        self._f_rho = cast_altitudes(f_geometric=False)(f_density)

        # Estimate the specific gas constant, gamma
        X = (("N2", 78.084), ("O2", 20.947_6), ("Ar", 0.934), ("CO2", 0.0314))
        X = ", ".join([f"{species}:{amount}" for (species, amount) in X])
        gas = GasModels.PerfectCaloric()
        gas.X = X
        self._f_gamma = lambda *args, **kwargs: gas.gamma

        return

    def _f_c_sound(self, altitude: Hint.nums,
                   geometric: bool = False) -> Quantity:
        """Return the speed of sound in air with altitude."""
        T = self.T(altitude, geometric)
        gamma = self.gamma(altitude, geometric)
        a = (gamma * co.STANDARD.ISO_2533.Rs * T) ** 0.5
        return a

    def _f_mu_visc(self, altitude: Hint.nums,
                   geometric: bool = False) -> Quantity:
        """Return the kinematic viscosity of air with altitude."""
        # Use Sutherland's law of viscosity
        T = self.T(altitude, geometric)
        mu = co.STANDARD.ISO_2533.betaS * T ** 1.5
        mu = mu / (T + co.STANDARD.ISO_2533.S)
        return mu


# ---------------------------------------------------------------------------- #


# U.S. Standard Atmosphere (1976)

class US1976(Atmosphere):
    """
    Simple model of the U.S. Standard Atmosphere (1976), as given in the
    NASA-TM-X-74335 specification.
    """

    def __init__(self, T_offset: float = None):
        # Super call
        super().__init__(T_offset=T_offset)

        raise NotImplementedError("Sorry! Coming soon\u2122")


# ---------------------------------------------------------------------------- #
# U.S. Department of Defense MIL-HDBK-310

milhdbk310_atms = dict()
for sheet, x_pct in itertools.product(list(df_MILHDBK310), ["1pct", "10pct"]):
    # Get fundamental atmopsheric properties (in terms of geometric alt.)
    table = cast2quantity(df_MILHDBK310[sheet])
    f_mil310_T, f_mil310_p, f_mil310_rho = atmosphere_builder(
        altitude=table["Geometric altitude, Z"],
        T=table[f"Temperature {x_pct}, T"],
        rho=table[f"Density {x_pct}, rho"],
        Mbar=co.STANDARD.ISO_2533.M
    )
    to_set_inside_init = dict([
        ("_f_T", cast_altitudes(f_geometric=True)(f_mil310_T)),
        ("_f_p", cast_altitudes(f_geometric=True)(f_mil310_p)),
        ("_f_rho", cast_altitudes(f_geometric=True)(f_mil310_rho)),
        ("_f_gamma", lambda *args, **kwargs: 1.4),  # Assume perfect gamma=1.4
        ("_f_c_sound", ISA1975().c_sound),  # Assume same method as ISA1975
        ("_f_mu_visc", ISA1975().mu_visc),  # Assume same method as ISA1975
        ("_f_k_thermal", ISA1975().k_thermal)  # Assume same method as ISA1975
    ])


    def _init(self, T_offset: float = None):
        """Dynamic initialisation function for MIL-HDBK-310 atmospheres."""
        # Super call
        Atmosphere.__init__(self, T_offset=T_offset)
        # We must set attributes like this so that the wrapper functions
        # properly (it needs to take self as an argument, be called inside init)
        for _, (k, v) in enumerate(to_set_inside_init.items()):
            setattr(self, k, v)
        return


    def _repr(self):
        """__repr__ string in the style of 'MILHDBK310_HD10_10pct'."""
        reprstring = f"{type(self).__name__}"
        return reprstring


    def _str(self):
        """__str__ string in the style of 'MILHDBK310[HD10:10%]'."""
        components = f"{type(self).__name__}".split("_")
        return f"{components[0]}[{components[1]}:{components[2][:-3]}%]"


    # Dynamically produce a class that permits access to the above properties
    attrs = dict([("__init__", _init), ("__repr__", _repr), ("__str__", _str)])
    model = type(f"MILHDBK310_{sheet}_{x_pct}", (Atmosphere,), attrs)
    milhdbk310_atms[f"{sheet}_{x_pct}"] = model(T_offset=0)  # Instantiate it!


# Doesn't matter what the name of the parent is, instantiate a catalogue of atms
class MILHDBK310(type("catalogue", (object,), milhdbk310_atms)):
    """A collection of atmospheric property models as given in MIL-HDBK-310."""

    def __init__(self):
        errormsg = (
            "This is a catalogue of atmospheres, and it should not be "
            "instantiated directly. Try one of my attributes!"
        )
        raise RuntimeError(errormsg)


# ---------------------------------------------------------------------------- #
# User "observed" atmospheres

class ObsAtmospherePerfect(Atmosphere):
    """
    Simple model for creating user-defined atmospheres based on observed
    atmospheric profiles (where the atmospheric gas components are assumed to
    behave ideally).
    """

    def __init__(self, altitude: Hint.nums = None, T: Hint.nums = None,
                 p: Hint.nums = None, rho: Hint.nums = None,
                 geometric: bool = None, Mbar: float = None,
                 T_offset: float = None):
        """
        Args:
            altitude: Sequence of altitudes.
            T: Sequence of temperatures. Optional, at least one of T,p, or rho
                is required.
            p: Sequence of pressures. Optional, at least one of T,p, or rho
                is required.
            rho: Sequence of densities. Optional, at least one of T,p, or rho
                is required.
            geometric: Whether or not the altitude sequence is geopotential or
                geometric in nature. Optional, defaults to False (geopotential).
            Mbar: The average molar mass of the atmosphere, per mol. Optional,
                defaults to that of the standard atmosphere.
            T_offset: Temperature profile offset from that in the temperature
                sequence. Optional, defaults to 0 offset.
        """
        # Super call
        super().__init__(T_offset=T_offset)

        # Recast as necessary
        geometric = False if geometric is None else geometric

        # Create functions of atmospheric properties with altitude
        f_temperature, f_pressure, f_density = atmosphere_builder(
            altitude=altitude, T=T, p=p, rho=rho, Mbar=Mbar)
        self._f_T = cast_altitudes(f_geometric=geometric)(f_temperature)
        self._f_p = cast_altitudes(f_geometric=geometric)(f_pressure)
        self._f_rho = cast_altitudes(f_geometric=geometric)(f_density)

        return
